import logging

from client import ScheduleCloudClient
from config.annotation import Value
from libs.utils import Singleton
from schedule import TaskThreadHolder

log = logging.getLogger('Ficus')

class ITaskQueueProxy(object):

    def count_waiting_tasks(self) -> int:
        pass


class LocalTaskQueueProxy(ITaskQueueProxy, Singleton):
    """
    本地模式的任务队列
    """

    def count_waiting_tasks(self) -> int:
        queue = TaskThreadHolder.count_trigger_queue()

        if len(queue) == 0:
            return 0

        result = 0
        for count in queue.values():
            result = result + count
        return result


class CeleryTaskQueueProxy(ITaskQueueProxy, Singleton):
    """
    云调度模式的任务队列
    """

    def count_waiting_tasks(self) -> int:
        # 2020-04-29 11:39:27 这里要修改一下, 更换成为调用 云调度的接口,而不是直接访问celery的redis,
        # 原因在于, redis的等待队列,在任务接收或执行的时候,就已经在队列中被移除了.这样的数量就不正确了.

        import config
        try:
            return ScheduleCloudClient.count_tasks(config.actor_name, status=["NEW", "RECEIVED", "STARTED"])
        except Exception as e:
            log.warn(f"count_waiting_tasks error:{str(e)}")
            return 0

        # import cloudcelery
        # import config
        # try:
        #     redis = cloudcelery.celery_redis_client()
        #     return redis.llen(config.actor_name)
        # except Exception as e:
        #     print(f"{str(e)}")
        #     return 0


class TaskQueueProxyFactory(Singleton):
    __instance = None

    @Value("${celery.enable:false}")
    def enable_celery(self):
        pass

    def get_task_queue_proxy(self) -> ITaskQueueProxy:
        if self.__instance:
            return self.__instance

        if self.enable_celery():
            # 使用celery 即云调度模式
            self.__instance = CeleryTaskQueueProxy.instance()
        else:
            # 使用本地模式
            self.__instance = LocalTaskQueueProxy.instance()

        return self.__instance


def task_queue_proxy() -> ITaskQueueProxy:
    return TaskQueueProxyFactory.instance().get_task_queue_proxy()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/23

from stomp.connect import StompConnection11

from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import Datasource
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from libs.utils import Singleton


class StompDatasourceDao(Singleton):

    @property
    def datasource(self) -> Datasource:
        data_source = get_multiple_datesource(FactDatasourceTypeEnum.JMS)
        return data_source

    @property
    def client(self) -> StompConnection11:
        return self.datasource.get_client()

    def send(self, destination, body, content_type=None, headers=None, **keyword_headers):
        """
        发送消息
        :param destination:
        :param body:
        :param content_type:
        :param headers:
        :param keyword_headers:
        :return:
        """
        self.client.send(destination, body, content_type, headers, **keyword_headers)



from api.exceptions import IllegalArgumentException, ServiceInnerException
from rule_engine import IRuleValueProvider
import re


class SimpleValueProvider(IRuleValueProvider):

    def __init__(self, data_type=None) -> None:
        self.__data_type = data_type

    def get_value(self, value):
        if value is None:
            return None
        elif self.__data_type is None:
            # 未知的类型,而且值又不为空,那么传进去的是什么 就是什么
            return value
        elif isinstance(value, str) and re.fullmatch(r"^\[([^\[^\].]+) TO ([^\[^\].]+)\]$", value):
            # 是范围的   '[10 TO 20]'
            # 表示是 如果是匹配的[XX TO XX]的格式,采用特殊方式解析
            aaa = re.findall(r"^\[([^\[^\].]+) TO ([^\[^\].]+)\]$", value)
            start_str = aaa[0][0]
            end_str = aaa[0][1]

            return [self.__convertValue(start_str,self.__data_type),self.__convertValue(end_str,self.__data_type)]
        elif isinstance(value, list) and len(value) == 1 and re.fullmatch(r"^([^\[^\].]+) TO ([^\[^\].]+)$", value[0]):
            # 同样是范围   list('10 TO 20')
            aaa = re.findall(r"^([^\[^\].]+) TO ([^\[^\].]+)$", value[0])
            start_str = aaa[0][0]
            end_str = aaa[0][1]

            return [self.__convertValue(start_str, self.__data_type), self.__convertValue(end_str, self.__data_type)]
        elif isinstance(value,list):
            # 是数组
            res = []
            for v in value:
                if isinstance(v,str):
                    res.append(self.__convertValue(v,self.__data_type))
                else:
                    res.append(v)
            return res
        elif isinstance(value,str) and re.fullmatch(r"^\[.+\]$", value):
            # 是数组字符串
            import json
            j = json.loads(value)
            res = []
            for v in j:
                if isinstance(v, str):
                    res.append(self.__convertValue(v, self.__data_type))
                else:
                    res.append(v)
            return res
        else:
            # 简单的值,直接转换
            return self.__convertValue(value,self.__data_type)

    def __convertValue(self,value,typeClass):
        if typeClass is None:
            raise IllegalArgumentException("类型转换错误,没有转换的目标类型")
        elif typeClass == object:
            # 这种也是没法转换的
            return value

        # 这有可能报转换异常
        try:
            import time
            import datetime
            from enum import Enum
            if typeClass in [time.struct_time,datetime.datetime]:
                # 日期特殊处理
                if typeClass == time.struct_time:
                    return time.strptime(value, "%Y-%m-%d")
                else:
                    return datetime.datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
            elif typeClass == Enum:
                return typeClass[value]
            elif typeClass == dict:
                import json
                return json.loads(value)
            else:
                return typeClass(value)
        except (TypeError,ValueError) as e:
            raise ServiceInnerException(f"字段值{value}失败.期望是:{typeClass}")


from api.model.MetaModelField import MetaModelField
from factdatasource.transformer.ITransformer import ITransformer

DATE_FORMAT_KEY = "__dateFormat__"
DEFAULT_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


class ParseDateTimeTransformer(ITransformer):
    """
    日期解析的转换器, 把字符串 转换成为 datetime
    """

    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField):

        if (data is None or len(data) == 0) or target_model_field.fieldName not in data:
            return {target_model_field.fieldName: None}

        # 能进来的只能是字符串
        old_value = data[target_model_field.fieldName]

        if old_value is None:
            return {target_model_field.fieldName: None}

        date_format_string = params[DATE_FORMAT_KEY] if DATE_FORMAT_KEY in params else DEFAULT_DATE_FORMAT

        from datetime import datetime
        return {target_model_field.fieldName: datetime.strptime(old_value, date_format_string)}



from api.model.MetaModelField import MetaModelField
from factdatasource.transformer.ITransformer import ITransformer


class FixItemTransformer(ITransformer):

    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField):
        # 如果没有这个字段就不管他

        if (data is None or len(data) == 0) or target_model_field.fieldName not in data:
            return {target_model_field.fieldName: None}

        # 能进来的只能是字符串
        old_value = data[target_model_field.fieldName]

        # 逻辑是这样的, 从params里面找到这个值得结果, 如果是 null 那么params中的key就是 null
        if old_value is None:
            return {target_model_field.fieldName: params["null"] if "null" in params else None}

        # 现在暂时先只支持 等于的方式的受控词. 以后再支持范围等方式的受控词

        return {target_model_field.fieldName: params[old_value] if old_value in params else old_value}



import datetime
import time
from abc import abstractmethod
from collections import Iterable
from enum import Enum
from typing import Iterator


# region 对象模型
class RuleActionTypeEnum(Enum):
    DEFAULT = 1     # 默认,直接返回 params
    MVEL = 2        # 用MVEL表达式 执行 action中的内容  PYTHON不支持
    BEAN = 3        # 从action中的字符串作为bean的源码,获取实现了 IAction的 源码  PYTHON不支持
    CLASS = 4       # 从action中的字符串作为类路径,反射获取 IAction实例  PYTHON不支持
    AVIATOR = 5     # 用aviator表达式 执行 action中的内容  PYTHON不支持
    PY_CLASS = 6    # python源码
    PY_BEAN = 7     # python类路径
    PY_EXP = 8      # python表达式  这里可能还需要一种 python表达式的类型,  因为 PY_BEAN或PY_CLASS其实都是需要实现IRuleAction接口的,应该需要一种表达式可以不限制输入的

    def actions(self):
        """
        返回默认实现
        :return:
        """
        if self in [RuleActionTypeEnum.MVEL,RuleActionTypeEnum.BEAN,RuleActionTypeEnum.CLASS,RuleActionTypeEnum.AVIATOR]:
            from rule_engine.action import UnSupportAction
            return UnSupportAction()
        elif  self == RuleActionTypeEnum.DEFAULT:
            from rule_engine.action import DefaultAction
            return DefaultAction()
        elif self == RuleActionTypeEnum.PY_EXP:
            from rule_engine.action import PyExpAction
            return PyExpAction()
        return None


class RuleConditionOperationEnum(Enum):
    DO_NOTHING = 0
    EQUAL = 1
    NOTEQUAL =2
    GT =3
    GTE =4
    LT =5
    LTE = 6
    IN = 7
    RANGE =8
    MVEL = 9        # 用MVEL表达式 执行 target 中的内容   PYTHON不支持
    BEAN = 10       # 从target中的字符串作为bean的源码,获取实现了 IRuleConditionOperation 的 bean  PYTHON不支持
    CLASS = 11      # 从target中的字符串作为类路径,反射获取 IRuleConditionOperation 实例 PYTHON不支持
    AVIATOR = 12    # 用aviator表达式 执行 target中的内容   PYTHON不支持
    PY_BEAN = 13     # python源码
    PY_CLASS = 14   # python类路径
    PY_EXP = 15     # python表达式  这里可能还需要一种 python表达式的类型,  因为 PY_BEAN或PY_CLASS其实都是需要实现IRuleConditionOperation接口的,应该需要一种表达式可以不限制输入的

    def allowed_target_type(self):
        """
        返回允许的字段类型
        :return:
        """
        if self in [RuleConditionOperationEnum.EQUAL,RuleConditionOperationEnum.NOTEQUAL,RuleConditionOperationEnum.IN]:
            return [int,float,str,time.struct_time,datetime.datetime,bool,Enum,dict,object]
        elif self in [RuleConditionOperationEnum.GT,RuleConditionOperationEnum.GTE,RuleConditionOperationEnum.LT,RuleConditionOperationEnum.LTE]:
            return [int,float,time.struct_time,datetime.datetime,object]
        elif self in [RuleConditionOperationEnum.RANGE]:
            return [int,float,object]

        return []

    def condition_operation(self):
        """
        返回类型的实例
        :return:
        """
        if self in [RuleConditionOperationEnum.MVEL,RuleConditionOperationEnum.BEAN,RuleConditionOperationEnum.CLASS,RuleConditionOperationEnum.AVIATOR]:
            from rule_engine.operation import UnSupportOperation
            return UnSupportOperation()
        elif self == RuleConditionOperationEnum.DO_NOTHING:
            from rule_engine.operation import DoNothingOperation
            return DoNothingOperation()
        elif self == RuleConditionOperationEnum.EQUAL:
            from rule_engine.operation import DefaultEqualOperation
            return DefaultEqualOperation()
        elif self == RuleConditionOperationEnum.NOTEQUAL:
            from rule_engine.operation import DefaultNotEqualOperation
            return DefaultNotEqualOperation()
        elif self == RuleConditionOperationEnum.IN:
            from rule_engine.operation import DefaultINOperation
            return DefaultINOperation()
        elif self == RuleConditionOperationEnum.GT:
            from rule_engine.operation import DefaultGTOperation
            return DefaultGTOperation()
        elif self == RuleConditionOperationEnum.GTE:
            from rule_engine.operation import DefaultGTEOperation
            return DefaultGTEOperation()
        elif self == RuleConditionOperationEnum.LT:
            from rule_engine.operation import DefaultLTOperation
            return DefaultLTOperation()
        elif self == RuleConditionOperationEnum.LTE:
            from rule_engine.operation import DefaultLTEOperation
            return DefaultLTEOperation()
        elif self == RuleConditionOperationEnum.RANGE:
            from rule_engine.operation import DefaultRangeOperation
            return DefaultRangeOperation()
        elif self == RuleConditionOperationEnum.PY_EXP:
            from rule_engine.operation import PyExpOperation
            return PyExpOperation()
        return None

class RuleConditionRelationEnum(Enum):
    """
    规则条件 相对于上一个条件的 逻辑关系, 默认是AND
    """
    AND = 1     #默认关系
    OR = 2

    def relation(self,one:bool,other:bool)->bool:
        if self == RuleConditionRelationEnum.OR:
            return one or other
        else:
            return one and other

class RuleConditionInfo:

    def __init__(self, source:str=None, operation:RuleConditionOperationEnum=None, rule_operation=None, target:str=None, target_provider=None, enable=True, subs:list=[], relation:RuleConditionRelationEnum=None) -> None:
        """

        :param source:
        :param operation:
        :param rule_operation:
        :param target:
        :param target_provider:
        :param enable:
        :param subs:
        :param relation:
        """
        self.source = source
        self.operation = operation
        self.rule_operation = rule_operation
        self.target = target
        self.target_provider = target_provider
        self.enable = enable
        self.subs = subs
        self.relation = relation or RuleConditionRelationEnum.AND

class RuleGroupInfo:
    """
    规则组对象
    """

    def __init__(self, name=None, project=None, match_all=False, enable=True, default_rule:str=None, rules=[]) -> None:
        self.name = name
        self.project = project
        self.match_all = match_all        # false表示 规则组中的规则具有排他性，满足某一规则后不再匹配组内其他规则.true 规则组中的规则无排他性，满足某一规则后继续匹配组内其他规则
        self.enable = enable        # 是否启用
        self.default_rule = default_rule      # 默认的规则ID
        self.rules = rules      # 规则 RuleInfo

class RuleAction:

    def __init__(self, action:str=None, rule_action=None, type:RuleActionTypeEnum=RuleActionTypeEnum.DEFAULT, params:dict = None) -> None:
        self.action = action
        self.rule_action = rule_action
        self.type = type
        self.params = params

class RuleInfo:

    def __init__(self,name=None,description=None,then=[],enable=True,when=[]) -> None:
        self.name = name
        self.description = description
        self.then = then or []
        self.enable = enable
        self.when = when or []

    def add_when(self, conditionInfo:RuleConditionInfo):
        self.when.append(conditionInfo)

    def add_then(self, action:RuleAction):
        self.then.append(action)

class RuleResult:

    def __init__(self, rule_name, success=True, result=[]) -> None:
        self.rule_name = rule_name
        self.success = success
        self.result = result or []

    def add_result(self, action):
        self.result.append(action)

    def __str__(self):
        return "{'rule_name':"+self.rule_name+",'success':"+str(self.success)+",result:"+str(self.result)+"}"

    def __repr__(self):
        return "{'rule_name':"+self.rule_name+",'success':"+str(self.success)+",result:"+str(self.result)+"}"

class RuleFacts(Iterable):
    """
    条件事实.key表示了一个规则因子的名字. value表示的是事实的值
    """

    def __init__(self,key:str=None,value=None,facts:dict=None) -> None:
        self.__facts = facts or {}
        if key is not None:
            self.add_fact(key,value)

    def add_fact(self, key:str, value):
        self.__facts[key] = value

    def add_facts(self, rule_facts):
        self.__facts.update(rule_facts.get_facts())

    def get_facts(self)->dict:
        return self.__facts

    def get_fact(self,key:str):
        # 这里需要使用python表达式或至少是点号表达式.  因为别个可能是 a.b.c 这样写的

        from munch import Munch
        dot_dic = Munch.fromDict(self.__facts)

        keys = key.split(".")

        result = None
        for k in keys:
            try:
                result = getattr(dot_dic,k)
                dot_dic = result if not isinstance(result,dict) else Munch.fromDict(result)
            except:
                return None

        if result is not None and isinstance(result,Munch):
            return result.toDict()

        return result

    def __iter__(self) -> Iterator:
        return self.__facts.__iter__()
# endregion

# region 接口方法
class IRuleAction:
    """
    注意,如果是在数据库里面写的是 $java:xxxxxxxxxxx 这种实例化出来的是单例的
    """

    @abstractmethod
    def do_action(self, action:str, facts:RuleFacts, params:dict):
        """

        :param action:
        :param facts:
        :param params:
        :return:
        """
        pass

class IRuleConditionOperation:
    """
    操作类型的接口定义.这个里面就决定了如何比较数值
    """
    @abstractmethod
    def do_operation(self, factValue:dict, targetValue)->bool:
        """
        对两个值进行比较,如果满足条件就返回true.否则返回false
        :param factValue:
        :param targetValue:
        :return:
        """
        pass

class IRuleListener:

    @abstractmethod
    def before_evaluate(self, rule, facts)->bool:
        """
        在一个规则比较前触发
        :param rule:
        :param facts:
        :return: true表示继续往下走,false表示强制取消这个规则的比较
        """
        pass

    @abstractmethod
    def after_evaluate(self, rule, facts, evaluation_result:bool):
        """
        比较后触发
        :param rule:
        :param facts:
        :param evaluation_result: 表示比较的结果
        :return:
        """
        pass

    @abstractmethod
    def before_action(self, rule, facts)->bool:
        """
        执行action动作前触发
        :param rule:
        :param facts:
        :return: true表示继续往下走,false表示不执行这个规则的action
        """
        pass

    @abstractmethod
    def on_success(self, rule, facts, result):
        """
        在action动作成功后触发
        :param rule:
        :param facts:
        :param result:
        :return:
        """
        pass

    @abstractmethod
    def on_failure(self, rule, facts, exception:Exception):
        """
        在action动作失败后触发
        :param rule:
        :param facts:
        :param exception:
        :return:
        """
        pass

class IRuleValueProvider:
    """
    数据值的提供者.他负责在运行时给各种规则提供数据
    使用这个接口的目的在于用户可以定制
    """

    @abstractmethod
    def get_value(self, value:str):
        """
        获取数据
        :param value:
        :return:
        """
# endregion


import logging

from rule_engine import RuleFacts, RuleGroupInfo
from rule_engine.core import IAgenda, PattenMatcher

log = logging.getLogger('Ficus')

class DefaultAgenda(IAgenda):


    def __init__(self, rule_group_info:RuleGroupInfo, rule_listeners:list) -> None:
        self.__rule_group_info = rule_group_info
        self.__rule_listeners = rule_listeners

    def get_matched_rules(self, rule_facts: RuleFacts) -> list:

        if not self.__rule_group_info.enable:
            log.info(f"规则组:{self.__rule_group_info.name}未启用,直接返回")
            return []

        # 1.取出所有的规则.开始遍历
        rules = self.__rule_group_info.rules

        # 是否匹配全部.如果是,就需要把每一个都遍历了.如果不是,只匹配到第一个符合条件的就返回
        is_match_all = self.__rule_group_info.match_all

        results = []

        for rule in rules:
            if not self.__triggerListenersBeforeEvaluate(rule, rule_facts):
                log.info(f"规则:{rule.name}在执行匹配前被忽略")
                continue

            if PattenMatcher.match(rule.when, rule_facts):
                # 如果匹配成功.就添加到结果中去
                self.__triggerListenersAfterEvaluate(rule, rule_facts, True)
                results.append(rule)
                if  not is_match_all:
                    # 匹配中一个就退出
                    break
            else:
                # 匹配失败
                self.__triggerListenersAfterEvaluate(rule, rule_facts, False)

        if len(results)==0:
            # 如果默认的不为空,并且是可使用的.就使用默认的
            if self.__rule_group_info.default_rule is not None and self.__rule_group_info.default_rule!="":
                for rule in rules:
                    if rule.name == self.__rule_group_info.default_rule:
                        results.append(rule)
                        log.info(f"规则组:{self.__rule_group_info.name}虽然没有匹配到规则,但是配置了默认规则{rule.name},命中默认规则")
                        break

        # 如果添加了默认的处理器还是空.就返回空
        return results

    # region 消息相关
    def __triggerListenersBeforeEvaluate(self, rule, facts):
        if self.__rule_listeners is not None:
            for listener in self.__rule_listeners:
                if not listener.before_evaluate(rule,facts):
                    return False
        return True

    def __triggerListenersAfterEvaluate(self, rule, facts, evaluation_result):
        if self.__rule_listeners is not None:
            for listener in self.__rule_listeners:
                listener.after_evaluate(rule, facts, evaluation_result)
    # endregion




#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import logging
from enum import Enum

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasource
from api.model.Page import Page
from factdatasource.dao.redis.RedisDatasourceDao import RedisDatasourceDao
from factdatasource.execptions import NotSupportedFDException, FDExecuteException
from factdatasource.dao.FactDatasource import customer_dao_context_holder as costomer_context
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext

log = logging.getLogger('Ficus')


class OptionTypeEnum(Enum):
    LIST = 'LIST'
    SET = 'SET'
    VALUE = 'VALUE'
    HASH = 'HASH'


class RedisFactDatasourceContext(AbstractFactDatasourceContext):
    """
    redis的target格式是 value:XXXXX:30 list:XXXX:30 set:XXXXX:30 hash:XXXXX:30
    要指明类型和过期时间(秒),如果没有指明类型默认是 value，没有指明过期时间默认不过期
    """

    def __init__(self, fact_datasource: FactDatasource):
        self.fact_datasource = fact_datasource
        self.target_type = None

    @property
    def dao(self) -> RedisDatasourceDao:
        return RedisDatasourceDao.instance()

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        source_name = self.fd().get_source_name()
        target_tuple = self.__analysis_target()
        option = target_tuple[0]
        name = target_tuple[1]
        try:
            costomer_context.set_source(source_name)
            if option == OptionTypeEnum.VALUE:
                result = self.dao.client.strlen(name)
            elif option == OptionTypeEnum.SET:
                result = self.dao.client.scard(name)
            elif option == OptionTypeEnum.HASH:
                result = self.dao.client.hlen(name)
            elif option == OptionTypeEnum.LIST:
                result = self.dao.client.llen(name)

            size = 0 if result is None else result
            return size
        except Exception as e:
            error = f'事实库{source_name}执行size操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        return self.size() <= 0

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        source_name = self.fd().get_source_name()
        target_tuple = self.__analysis_target()
        option = target_tuple[0]
        name = target_tuple[1]
        try:
            costomer_context.set_source(source_name)
            if option == OptionTypeEnum.VALUE:
                result = [self.dao.client.get(name)]
            elif option == OptionTypeEnum.SET:
                result = self.dao.client.srandmember(name, size)
            elif option == OptionTypeEnum.HASH:
                result = [self.dao.client.hgetall(name)]
            elif option == OptionTypeEnum.LIST:
                result = self.dao.client.lrange(name, 0, size)

            if result and size and len(result) > size:
                return result[0:size]
            return result
        except Exception as e:
            error = f'事实库{source_name}执行collect操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        raise NotSupportedFDException('redis不支持该操作')

    def query(self, query: str, parameters: dict = None):
        """
        使用查询语句查询数据
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        source_name = self.fd().get_source_name()
        target_tuple = self.__analysis_target()
        option = target_tuple[0]
        name = target_tuple[1]

        page = None
        if parameters and 'pageNum_' in parameters.keys() and 'pageSize_' in parameters.keys():
            # 需要分页的
            param = {
                'pageNum': parameters.get('pageNum_'),
                'pageSize': parameters.get('pageSize_'),
                'needCount': parameters.get('needCount_') if parameters.get('needCount_') else False
            }
            page = Page(**param)

        try:
            costomer_context.set_source(source_name)
            if option == OptionTypeEnum.VALUE:
                # value不会进行分页
                result = [self.dao.client.get(name)]
                page = Page.no_page(result)
            elif option == OptionTypeEnum.SET:
                result = list(self.dao.client.smembers(name))
                if page:
                    tmp = result[page.start_row:page.end_row]
                    page.set_result(tmp, len(result))
                else:
                    # TODO 这里如果没有分页，数据量过大可能会有问题
                    page = Page.no_page(result)
            elif option == OptionTypeEnum.HASH:
                if not query or '*' == query:
                    # 不能分页，顺序不好保证
                    result = [self.dao.client.hgetall(name)]
                else:
                    result = [self.dao.client.hget(name, query)]
                page = Page.no_page(result)
            elif option == OptionTypeEnum.LIST:
                if page:
                    result = self.dao.client.lrange(name, page.start_row, page.end_row)
                    total = None
                    if page.need_count():
                        total = self.dao.client.llen(name)
                    page.set_result(result, total)
                else:
                    result = self.dao.client.lrange(name, 0, -1)
                    page = Page.no_page(result)
            return page
        except Exception as e:
            error = f'事实库{source_name}执行size操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def inserts(self, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        if not result_list:
            return
        if not isinstance(result_list, list):
            result_list = [result_list]

        source_name = self.fd().get_source_name()
        target_tuple = self.__analysis_target()
        option = target_tuple[0]
        name = target_tuple[1]
        expire = target_tuple[2]

        try:
            costomer_context.set_source(source_name)
            if option == OptionTypeEnum.VALUE:
                for result in result_list:
                    self.dao.client.set(name, result)
            elif option == OptionTypeEnum.SET:
                self.dao.client.sadd(name, *result_list)
            elif option == OptionTypeEnum.HASH:
                for result in result_list:
                    self.dao.client.hmset(name, result)
            elif option == OptionTypeEnum.LIST:
                self.dao.client.rpush(name, *result_list)

            # 设置超时时间
            if expire:
                self.dao.client.expire(name, expire)
        except Exception as e:
            error = f'事实库{source_name}执行size操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def updates(self, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        self.inserts(result_list)

    def inserts_or_updates(self, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        self.inserts(result_list)

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        # redis已经很快了,暂时不使用并发的方式执行
        raise NotSupportedFDException('redis不支持该操作')

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        # redis已经很快了,暂时不使用并发的方式执行
        raise NotSupportedFDException('redis不支持该操作')

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        # redis已经很快了,暂时不使用并发的方式执行
        raise NotSupportedFDException('redis不支持该操作')

    def delete_all(self):
        """
        清空数据
        :return:
        """
        source_name = self.fd().get_source_name()
        target_tuple = self.__analysis_target()
        name = target_tuple[1]
        try:
            costomer_context.set_source(source_name)
            self.dao.client.delete(name)
        except Exception as e:
            error = f'事实库{source_name}执行delete_all操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :return:
        """
        raise NotSupportedFDException('redis不支持该操作')

    def delete_conditions(self, condition_groups: list):
        raise NotSupportedFDException('redis不支持该操作')

    def __analysis_target(self) -> tuple:
        """
        分析redis的target，获取配置的类型和过期时间
        redis的 目标格式是  value:XXXXX:30  list:XXXX:30 set:XXXXX:30 hash:XXXXX:30
        需要指明类型.和过期时间 如果没有指明类型默认是 value 默认不过期
        :return: (OptionTypeEnum, target, int)
        """
        if self.target_type:
            return self.target_type

        target = self.fd().target
        if not target:
            raise IllegalArgumentException(f'redis的target配置错误，target不能为空。')

        split = target.split(":", 2)

        def get_type(option: str):
            if not option:
                return OptionTypeEnum.VALUE
            tmp = option.upper()
            if tmp == 'LIST':
                return OptionTypeEnum.LIST
            elif tmp == 'SET':
                return OptionTypeEnum.SET
            elif tmp == 'VALUE':
                return OptionTypeEnum.VALUE
            elif tmp == 'HASH':
                return OptionTypeEnum.HASH
            else:
                return OptionTypeEnum.VALUE

        if len(split) == 3:
            # 有超时时间
            self.target_type = (get_type(split[0]), split[1], int(split[2]), )
        elif len(split) == 2:
            # 没有超时时间
            self.target_type = (get_type(split[0]), split[1], None, )
        else:
            # 只有队列名
            self.target_type = (OptionTypeEnum.VALUE, split[0], None, )
        return self.target_type









#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from urllib.parse import quote_plus

from pymongo import MongoClient
from pymongo.database import Database

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import DatasourceListener, MultipleBaseDatasource, BaseDatasource
from factdatasource.execptions import DatasourceNotFoundException, FDExecuteException
from libs.utils import Singleton


class MultipleMongoDatasource(DatasourceListener, MultipleBaseDatasource, Singleton):
    """
    管理fd中的所有MONGO类型的数据源
    """

    def __init__(self):
        # source_name --> MongoDatasource
        self._target_dataSources = dict()

    def get_datasource_type(self):
        """
        获取数据源类型
        :return:
        """
        return FactDatasourceTypeEnum.MONGO

    def add_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        添加一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'修改数据源参数错误:source_name、url都不能为空')

        for target in self._target_dataSources.values():
            if target.url == url and target.credentials == credentials:
                # url 和 credentials 都相同，说明是同一个数据库连接就不再重复创建数据源了
                target.add_source_name(source_name)
                self._target_dataSources[source_name] = target
                return None

        # 开始创建数据源
        mongo_data_source = MongoDatasource(source_name, url, credentials)
        mongo_data_source.start()
        self._target_dataSources[source_name] = mongo_data_source

    def update_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        修改一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """
        if not source_name or not url:
            raise IllegalArgumentException(f'修改数据源参数错误:source_name、url都不能为空')

        self.delete_datasource_type(source_name)
        self.add_datasource_type(source_name, url, credentials)

    def delete_datasource_type(self, source_name: str):
        """
        删除一个数据源
        :param source_name:
        :return:
        """
        if not source_name:
            raise IllegalArgumentException(f'删除数据源参数错误:source_name不能为空')

        target: MongoDatasource = self._target_dataSources.pop(source_name)
        if target:
            if target.only_one_source():
                target.close_client()
            else:
                target.remove_source_name(source_name)

    def get_data_source(self):
        """
        获取数据源的基本信息
        :return:
        """
        source_name = self.determine_current_lookup_key()
        if not source_name:
            raise FDExecuteException('未设置操作源,无法获取数据源信息。')

        if source_name in self._target_dataSources.keys():
            return self._target_dataSources[source_name]
        raise DatasourceNotFoundException('未发现数据源%s,请添加后再获取。' % source_name)

    def get_client(self):
        """
        获取数据库操作的client
        :return:
        """
        target: MongoDatasource = self.get_data_source()
        return target.get_client()

    def close_client(self):
        """
        关闭客户端，关闭连接
        :return:
        """
        target: MongoDatasource = self.get_data_source()
        return target.close_client()


class MongoDatasource(BaseDatasource):

    def __init__(self, source_name, url, credentials):
        self.init(source_name, url, credentials)
        # 使用的库
        self.db = None
        # 连接mongo的uri
        self.uri = None
        self.client: MongoClient = None

    def start(self):
        """
        :return:
        """
        self._parse_url()
        self.client = MongoClient(host=self.uri, maxPoolSize=20, connect=False)

    def get_client(self) -> Database:
        return self.client[self.db]

    def close_client(self):
        if self.client:
            self.client.close()

    def _parse_url(self):
        """
        分析url，构造连接mongo的uri
        :return:
        """
        if self.credentials:
            if '@' in self.credentials:
                con_credential = self.credentials[0:self.credentials.find('@')]
                user_db = self.credentials[self.credentials.find('@')+1:]
            else:
                con_credential = self.credentials
                user_db = "admin"
        else:
            con_credential = ''
            user_db = "admin"

        if con_credential and ':' in con_credential:
            uri_credential = quote_plus(con_credential[0:con_credential.find(':')]) + ':' \
                             + quote_plus(con_credential[con_credential.find(':')+1:]) + '@'
        else:
            uri_credential = ''

        if not self.url or '@' not in self.url:
            raise IllegalArgumentException(f'mongo连接信息配置错误，无法进行连接：{self.url}')

        con_url = self.url[0:self.url.find('@')]
        db = self.url[self.url.find('@')+1:]
        if not db:
            raise IllegalArgumentException(f'mongo连接信息未配置使用库，无法进行连接：{self.url}')

        # mongodb://sdba:sdba@localhost:27017,localhost:27018/admin
        _uri = f'mongodb://{uri_credential}{con_url}/{user_db}'

        self.db = db
        self.uri = _uri
        return _uri



from queue import Queue

from api.exceptions import ServiceInnerException
from api.model import OutputWrapper
from api.model.ReadWriteLock import ReadWriteLock


class OtherTargetOutputPipe:
    __index = 1
    __output_stream: Queue = None
    __output_count = 0
    __lock = ReadWriteLock()

    def __init__(self, output_stream, index, output_count, lock):
        self.__output_stream = output_stream
        self.__lock = lock
        self.__index = index
        self.__output_count = output_count

    def then_output_for_insert(self, callable_func):
        # 必须保证callable不为空, 并且 当前的输出层级 小于输出目标的个数,才能执行输出.
        # 比如outputFdCode配置的是:["a","b"]两个. 那么outputCount = 2. 第一次进入thenOutputForInsert的时候index=1 满足要求,允许输出.
        # 如果这个时候再调用 thenOutputForInsert index=2 那么就不会再执行callable里面的东西了,因为就算执行了,也没有对应的outputfd提供输出.
        if callable_func is not None and self.__output_count > self.__index:
            self.__lock.acquire_read()
            try:
                contents = callable_func()
                if contents is not None:
                    if isinstance(contents, list):
                        for content in contents:
                            self.__output_stream.put_nowait(OutputWrapper.INSERT(content, self.__index))
                    else:
                        self.__output_stream.put_nowait(OutputWrapper.INSERT(contents, self.__index))
            except Exception as e:
                raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
            finally:
                self.__lock.release_read()

        self.__index = self.__index + 1
        return self

    def then_flat_output_for_updates(self, callable_func):
        # 必须保证callable不为空, 并且 当前的输出层级 小于输出目标的个数,才能执行输出.
        # 比如outputFdCode配置的是:["a","b"]两个. 那么outputCount = 2. 第一次进入thenOutputForInsert的时候index=1 满足要求,允许输出.
        # 如果这个时候再调用 thenOutputForInsert index=2 那么就不会再执行callable里面的东西了,因为就算执行了,也没有对应的outputfd提供输出.
        if callable_func is not None and self.__output_count > self.__index:
            self.__lock.acquire_read()
            try:
                contents = callable_func()
                if contents is not None:
                    if isinstance(contents, list):
                        for content in contents:
                            self.__output_stream.put_nowait(OutputWrapper.UPDATE(content, self.__index))
                    else:
                        self.__output_stream.put_nowait(OutputWrapper.UPDATE(contents, self.__index))
            except Exception as e:
                raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
            finally:
                self.__lock.release_read()

        self.__index = self.__index + 1
        return self

    def then_flat_output_for_upserts(self, callable_func):
        # 必须保证callable不为空, 并且 当前的输出层级 小于输出目标的个数,才能执行输出.
        # 比如outputFdCode配置的是:["a","b"]两个. 那么outputCount = 2. 第一次进入thenOutputForInsert的时候index=1 满足要求,允许输出.
        # 如果这个时候再调用 thenOutputForInsert index=2 那么就不会再执行callable里面的东西了,因为就算执行了,也没有对应的outputfd提供输出.
        if callable_func is not None and self.__output_count > self.__index:
            self.__lock.acquire_read()
            try:
                contents = callable_func()
                if contents is not None:
                    if isinstance(contents, list):
                        for content in contents:
                            self.__output_stream.put_nowait(OutputWrapper.UPSERT(content, self.__index))
                    else:
                        self.__output_stream.put_nowait(OutputWrapper.UPSERT(contents, self.__index))
            except Exception as e:
                raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
            finally:
                self.__lock.release_read()

        self.__index = self.__index + 1
        return self

    def then_flat_output_for_delete(self, callable_func):
        # 必须保证callable不为空, 并且 当前的输出层级 小于输出目标的个数,才能执行输出.
        # 比如outputFdCode配置的是:["a","b"]两个. 那么outputCount = 2. 第一次进入thenOutputForInsert的时候index=1 满足要求,允许输出.
        # 如果这个时候再调用 thenOutputForInsert index=2 那么就不会再执行callable里面的东西了,因为就算执行了,也没有对应的outputfd提供输出.
        if callable_func is not None and self.__output_count > self.__index:
            self.__lock.acquire_read()
            try:
                contents = callable_func()
                if contents is not None:
                    if isinstance(contents, list):
                        for content in contents:
                            self.__output_stream.put_nowait(OutputWrapper.DELETE(content, self.__index))
                    else:
                        self.__output_stream.put_nowait(OutputWrapper.DELETE(contents, self.__index))
            except Exception as e:
                raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
            finally:
                self.__lock.release_read()

        self.__index = self.__index + 1
        return self

    def then_clean(self):
        # 必须保证callable不为空, 并且 当前的输出层级 小于输出目标的个数,才能执行输出.
        # 比如outputFdCode配置的是:["a","b"]两个. 那么outputCount = 2. 第一次进入thenOutputForInsert的时候index=1 满足要求,允许输出.
        # 如果这个时候再调用 thenOutputForInsert index=2 那么就不会再执行callable里面的东西了,因为就算执行了,也没有对应的outputfd提供输出.
        if self.__output_count > self.__index:
            self.__lock.acquire_read()
            try:
                self.__output_stream.put_nowait(OutputWrapper.CLEAN(self.__index))
            except Exception as e:
                raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
            finally:
                self.__lock.release_read()

        self.__index = self.__index + 1
        return self

    def then_skip_output(self):
        self.__index = self.__index + 1
        return self



from api.handler.ICacheAble import ICacheAble
from api.handler.script import ScriptHandlerHolder


class ICacheAbleScript(ICacheAble):

    def set_cache_value(self, key, value):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            handler.set_cache_value(key,value)

    def set_cache_value_if_absent(self, key, value):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            return handler.set_cache_value_if_absent(key, value)
        return False

    def get_cache_value(self, key):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            return handler.get_cache_value(key)
        return None

    def delete_cache_value(self, key):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            handler.delete_cache_value(key)

    def set_cache_value_from_process(self, key, value):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            handler.set_cache_value_from_process(key, value)

    def set_cache_value_if_absent_from_process(self, key, value):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            return handler.set_cache_value_if_absent_from_process(key, value)
        return False

    def get_cache_value_from_process(self, key):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            return handler.get_cache_value_from_process(key)
        return None

    def delete_cache_value_from_process(self, key):
        handler = ScriptHandlerHolder.cacheHandler()
        if handler is not None:
            handler.delete_cache_value_from_process(key)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/7
import datetime

import pytz

from api.annotation.annotation import TaskHandler
from api.handler.ce.ISimpleScriptCE import ISimpleScriptCE
from api.handler.crawl.ISimpleScriptCrawl import ISimpleScriptCrawl
from api.model.FactDatasource import FactDatasource
from api.model.FdInputPipe import FdInputPipe
from schedule import TaskHandlerContext
from schedule.test.suite import DataComputeExecution
from schedule.test.suite.ce import AbstractSimpleScriptCETestSuite


@TaskHandler("demoSimpleScriptCE")
class DemoSimpleScriptCE(ISimpleScriptCE):
    """CE_python脚本"""
    def do_compute(self, source_fds: FdInputPipe, params: dict):
        date = datetime.datetime.now(pytz.timezone('Asia/Shanghai')).strftime('%Y-%m-%d %H:%M:%S')
        result = {
            "id": 4,
            "tv_name": "CCTV-Python-CE",
            "epg_name": "Python-CE生财有道",
            "rank": 60,
            "date": date
        }
        if params and params.get("name"):
            name = params.get("name")
            result["name"] = name
            self.task_logger.log(f"CE_PYTHON日志测试，接收到环境变量name:{name}")

        self.task_logger.log("CE_PYTHON日志测试，这里输出普通日志")
        try:
            raise Exception('CE_PYTHON日志测试，这里输出错误日志')
        except Exception as e:
            self.task_logger.error(e)

        query_sql = """{"from": 0,"size": 200}"""
        query_result = source_fds.get_fd('test_source').query(query_sql)
        self.task_logger.log(f"CE_PYTHON日志测试，查询结果:{query_result}")

        # 缓存
        self.set_cache_value('name', 'sobey')
        cache_value = self.get_cache_value('name')
        self.task_logger.log(f"缓存值{cache_value}")

        from api.model import OutputWrapper
        return [OutputWrapper.UPSERT(result)]

    def kill(self):
        self.task_logger.log("CE_PYTHON脚本被终止")


class TestAbstractSimpleScriptCETestSuite(AbstractSimpleScriptCETestSuite):

    def mock_target(self) -> ISimpleScriptCrawl:
        return TaskHandlerContext.load_task_handler("demoSimpleScriptCE")

    def mock_fact_datasource(self) -> dict:
        fd = {
            "code": "test_source",
            "connection": "localhost:9200",
            "credentials": None,
            "model": None,
            "name": "test_source",
            "privilege": None,
            "projectCode": "TEST",
            "site": "S1",
            "tags": [
                "masterData"
            ],
            "target": "es_test",
            "ttl": 0,
            "type": "ES"
        }

        fds = {
            'test_source': FactDatasource(**fd)
        }
        return fds

    def mock_ce(self) -> DataComputeExecution:
        dc = {
            'site': 'S1',
            'projectCode': 'TEST',
            'code': 'test_script_crawl',
            'sourceFdCodes': ['test_source'],
            'outputFdCodes': ['test_source'],
            'params': {
                'name': 'sobey'
            }
        }
        return DataComputeExecution(**dc)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
from abc import abstractmethod

import requests
from munch import Munch

from api.exceptions import ServiceInnerException, IllegalArgumentException, AuthException
from client import check_instance_avaliable, do_service
from config.annotation import Value
from factdatasource.FactDatasourceContext import FactDatasourceContext
from factdatasource.FactDatasourceContextHolder import FactDatasourceContextHolder
from factdatasource.dao.jdbc.MultipleJdbcDatasource import register_jdbc_default_datasource
from factdatasource.dao.kafka.MultipleKafkaDatasource import register_kafka_default_datasource
from factdatasource.message.FactDatasourceMessageHanlder import FactDatasourceChangeListener
from libs.utils import Singleton
from service.FactDatasourceService import FactDatasourceService


class FactDatasourceProxy(object):

    @abstractmethod
    def fd(self, fd_code: str):
        """
        返回fd对象
        :param fd_code:
        :return: FactDatasource
        """

    @abstractmethod
    def size(self, fd_code: str):
        """
        返回数据总长度
        :param fd_code:
        :return: 数据条数:long
        """

    @abstractmethod
    def is_empty(self, fd_code: str):
        """
        返回是否存在数据
        :param fd_code:
        :return: boolean
        """

    @abstractmethod
    def collect(self, fd_code: str, offset: int, size: int, only_model_field: bool):
        """
        返回指定条数的数据
        :param offset:
        :param only_model_field:
        :param fd_code:
        :param size: 返回的条数
        :return: list
        """

    @abstractmethod
    def collect_conditions(self, fd_code: str, offset: int, size: int, condition_groups: list, only_model_field: bool):
        """
        返回指定查询条件的数据
        :param offset:
        :param only_model_field:
        :param fd_code:
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """

    @abstractmethod
    def query(self, fd_code: str, query: str, parameters: dict):
        """
        使用查询语句查询数据
        :param fd_code:
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """

    @abstractmethod
    def inserts(self, fd_code: str, result_list: list) -> list:
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param fd_code:
        :param result_list: 要保存的数据
        :return:
        """

    @abstractmethod
    def updates(self, fd_code: str, result_list: list) -> list:
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param fd_code:
        :param result_list: 要修改的数据
        :return:
        """

    @abstractmethod
    def save_or_updates(self, fd_code: str, result_list: list) -> list:
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param fd_code:
        :param result_list: 要添加或者需要修改的数据
        :return:
        """

    @abstractmethod
    def delete_all(self, fd_code: str):
        """
        清空数据
        :param fd_code:
        :return:
        """

    @abstractmethod
    def delete(self, fd_code: str, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :param fd_code:
        :param query:
        :return:
        """

    @abstractmethod
    def delete_conditions(self, fd_code: str, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param fd_code:
        :param condition_groups:
        :return:
        """

    # @abstractmethod
    # def get_fact_datasource_fields(self, fd_code: str):
    #     """
    #     这个python版暂时用不到，先不实现
    #     获取fd的字段
    #     :param fd_code:
    #     :return: List<FactDatasourceField>
    #     """

    @abstractmethod
    def exists_fds(self, fd_codes: list) -> bool:
        """
        判断fd是否存在
        :param fd_codes:
        :return:
        """


class DistributedFactDatasourceProxy(FactDatasourceProxy, Singleton):
    """
    本地直接使用FD操作数据库的方式
    """

    def fd(self, fd_code: str):
        """
        返回fd对象
        :param fd_code:
        :return: FactDatasource
        """
        return self._get_fd_context(fd_code).fd()

    def size(self, fd_code: str):
        """
        返回数据总长度
        :param fd_code:
        :return: 数据条数:long
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'read')
        return fd_context.size()

    def is_empty(self, fd_code: str):
        """
        返回是否存在数据
        :param fd_code:
        :return: boolean
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'read')
        return fd_context.is_empty()

    def collect(self, fd_code: str, offset: int, size: int, only_model_field: bool = False):
        """
        返回指定条数的数据
        :param offset:
        :param only_model_field:
        :param fd_code:
        :param size: 返回的条数
        :return: list
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'read')
        result = fd_context.collect(size)
        # 这里需要对查询结果做字段过滤
        self._inner_field_permission(fd_context, result, only_model_field)
        return result

    def collect_conditions(self, fd_code: str, offset: int, size: int, condition_groups: list, only_model_field: bool = False):
        """
        返回指定条数的数据
        :param offset:
        :param only_model_field:
        :param fd_code:
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'read')
        result = fd_context.collect_conditions(size, condition_groups)
        # 这里需要对查询结果做字段过滤
        self._inner_field_permission(fd_context, result, only_model_field)
        return result

    def query(self, fd_code: str, query: str, parameters: dict):
        """
        使用查询语句查询数据
        :param fd_code:
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'read')
        result = fd_context.query(query, parameters)
        # 这里需要对查询结果做字段过滤
        self._inner_field_permission(fd_context, result, False)
        return result

    def inserts(self, fd_code: str, result_list: list) -> list:
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param fd_code:
        :param result_list: 要保存的数据
        :return:
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'write')
        return fd_context.inserts(result_list)

    def updates(self, fd_code: str, result_list: list) -> list:
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param fd_code:
        :param result_list: 要修改的数据
        :return:
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'write')
        return fd_context.updates(result_list)

    def save_or_updates(self, fd_code: str, result_list: list) -> list:
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param fd_code:
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'write')
        return fd_context.inserts_or_updates(result_list)

    def delete_all(self, fd_code: str):
        """
        清空数据
        :param fd_code:
        :return:
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'delete')
        return fd_context.delete_all()

    def delete(self, fd_code: str, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :param fd_code:
        :param query:
        :return:
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'delete')
        return fd_context.delete(query)

    def delete_conditions(self, fd_code: str, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param fd_code:
        :param condition_groups: 传入的是 ConditionGroup对象
        :return:
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'delete')
        return fd_context.delete_conditions(condition_groups)

    # def get_fact_datasource_fields(self, fd_code: str):
    #     """
    #     获取fd的字段
    #     :param fd_code:
    #     :return: List<FactDatasourceField>
    #     """
    #     fd_context = self._get_fd_context(fd_code)
    #     self._inner_check_permission(fd_context, 'execute')
    #     return fd_context.get_fact_datasource_fields()

    def exists_fds(self, fd_codes: list) -> bool:
        """
        判断fd是否存在
        :param fd_codes:
        :return:
        """
        if not fd_codes:
            return False

        if not isinstance(fd_codes, list):
            raise IllegalArgumentException("检测fd是否存在失败,输入参数不是一个list")

        fd_service = FactDatasourceService.instance()
        return fd_service.exists_fds(fd_codes)

    def _get_fd_context(self, fd_code: str):
        fd_context: FactDatasourceContext = FactDatasourceContextHolder.instance().get_fact_datasource(fd_code)
        return fd_context

    def _inner_check_permission(self, fd_context: FactDatasourceContext, operation: str):
        # TODO 暂未实现内容权限验证
        return True

    def _inner_field_permission(self, fd_context, result, only_model_field):
        # TODO 暂未实现内容权限验证及只返回模型有的字段
        return True


class CentralizedFactDatasourceProxy(FactDatasourceProxy, Singleton):
    """
    集中式的FD获取,也就是传统的 调用ficus-web的方式来获取fd
    """

    def fd(self, fd_code: str):
        """
        获取某一个FD对象
        :param fd_code: fd的唯一code
        :return: FD对象
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}")

            if r is not None:
                return Munch(r)
            else:
                return None
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def size(self, fd_code: str):
        """
        获取数据的条数
        :param fd_code:  fd的唯一code
        :return: 数据的条数 没得就返回0
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/size")

            if r is not None:
                return r
            else:
                return 0
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def is_empty(self, fd_code: str):
        """
        判断数据集是否为空
        :param fd_code: fd的唯一code
        :return: 如果为空返回True,否则返回False
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/empty")

            if r is not None:
                return r
            else:
                return True
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def collect(self, fd_code: str, offset: int, size: int, only_model_field: bool = False):
        """
        返回整个数据集
        :param offset:
        :param only_model_field:是否只返回模型里面的字段
        :param size:
        :param fd_code: fd的唯一code
        :return:
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/all", params={'offset': offset, 'size': size, 'onlyModelField': only_model_field})
            if r is not None:
                if isinstance(r, list):
                    return [Munch(x) for x in r]
                else:
                    return Munch(r)
            else:
                return None
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def collect_conditions(self, fd_code: str, offset: int, size: int, condition_groups: list, only_model_field: bool = False):
        """
        返回指定条数的数据
        :param offset:
        :param only_model_field:
        :param fd_code:
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/conditions", method="post", data=condition_groups,
                           params={'offset': offset, 'size': size, 'onlyModelField': only_model_field})
            if r is not None:
                if isinstance(r, list):
                    return [Munch(x) for x in r]
                else:
                    return Munch(r)
            else:
                return None
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def query(self, fd_code: str, query: str, parameters: dict):
        """
        对数据集进行查询
        :param fd_code: fd的唯一code
        :param query_str: 查询语句
        :param parameters: 查询可能涉及的参数 K/V形式
        :return:
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/query", method="post", data=parameters,
                           params={'query': query})

            if r is not None:
                if isinstance(r, list):
                    return [Munch(x) for x in r]
                else:
                    return Munch(r)
            else:
                return None
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def inserts(self, fd_code: str, result_list: list) -> list:
        """
        增加数据
        :param fd_code: fd的唯一code
        :param result_list: 要被写入的数据 是一个List
        :return:
        """
        if result_list is None:
            return None

        if not isinstance(result_list, list):
            raise IllegalArgumentException("输入的参数:result_list 不是数组")

        check_instance_avaliable()

        request = []
        for result in result_list:
            if not isinstance(result, Munch):
                # 这里所有类型都要添加进去
                request.append(result)
            else:
                # 说明是munch的,那么就转成Dict的
                request.append(result.toDict())

        try:
            r = do_service(f"/remote/fd-service/{fd_code}", method="post", data=request, return_type="json")
            if r is not None:
                if isinstance(r, list):
                    return [Munch(x) for x in r]
                else:
                    return None
            else:
                return None
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def updates(self, fd_code: str, result_list: list) -> list:
        """
        更新数据
        :param fd_code: fd的唯一code
        :param result_list: 要被更新的数据 是一个List
        :return:
        """
        if result_list is None:
            return None

        if not isinstance(result_list, list):
            raise IllegalArgumentException("输入的参数:result_list 不是数组")

        check_instance_avaliable()

        request = []
        for result in result_list:
            if not isinstance(result, Munch):
                # 这里所有类型都要添加进去
                request.append(result)
            else:
                # 说明是munch的,那么就转成Dict的
                request.append(result.toDict())

        try:
            r = do_service(f"/remote/fd-service/{fd_code}", method="put", data=request, return_type="json")
            if r is not None:
                if isinstance(r, list):
                    return [Munch(x) for x in r]
                else:
                    return None
            else:
                return None
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def save_or_updates(self, fd_code: str, result_list: list) -> list:
        """
        新增或更新数据
        :param fd_code:  fd的唯一code
        :param result_list: upsertCache
        :return:
        """
        if result_list is None:
            return None

        if not isinstance(result_list, list):
            raise IllegalArgumentException("输入的参数:result_list 不是数组")

        check_instance_avaliable()

        request = []
        for result in result_list:
            if not isinstance(result, Munch):
                # 这里所有类型都要添加进去
                request.append(result)
            else:
                # 说明是munch的,那么就转成Dict的
                request.append(result.toDict())

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/upsert", method="post", data=request, return_type="json")
            if r is not None:
                if isinstance(r, list):
                    return [Munch(x) for x in r]
                else:
                    return None
            else:
                return None
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def delete_all(self, fd_code: str):
        """
        删除数据集中所有数据
        :param fd_code: fd的唯一code
        :return:
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}", method="delete", return_type="None")
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def delete(self, fd_code: str, query: str):
        """
        删除数据集中查询语句命中的数据
        :param fd_code: fd的唯一code
        :param query_str: 查询语句
        :return:
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/query", method="delete", params={'query': query},
                           return_type="None")
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def delete_conditions(self, fd_code: str, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param fd_code:
        :param condition_groups:
        :return:
        """
        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/{fd_code}/conditions", method="delete", data=condition_groups,
                           return_type="None")
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e

    def get_fact_datasource_fields(self, fd_code: str):
        """
        获取fd的字段
        :param fd_code:
        :return: List<FactDatasourceField>
        """
        fd_context = self._get_fd_context(fd_code)
        self._inner_check_permission(fd_context, 'execute')
        return fd_context.get_fact_datasource_fields()

    def exists_fds(self, fds):
        """
        判断fd是否存在
        :param fds:
        :return:
        """
        if fds is None:
            return False

        if not isinstance(fds, list):
            raise IllegalArgumentException("检测fd是否存在失败,输入参数不是一个list")

        check_instance_avaliable()

        try:
            r = do_service(f"/remote/fd-service/exists", method="post", data=fds)
            if r is not None:
                return r
            else:
                return False
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 500:
                # 说明服务器端报错了
                raise ServiceInnerException(e.response._content.decode('utf-8'))
            elif e.response.status_code >= 400 and e.response.status_code < 500 and e.response.status_code != 404:
                # 说明是认证相关的错误
                raise AuthException(f"服务端异常 {str(e)} {e.response._content.decode('utf-8')}")
            raise e


@Value("${sobeycube.factdatasource.distributed-mode:False}")
def distributed_mode():
    pass


class FactDatasourceProxyFactory(Singleton):
    __instance = None

    def get_fd_client_proxy(self, mode: bool = None) -> FactDatasourceProxy:
        if self.__instance:
            return self.__instance

        if mode is None:
            mode = distributed_mode()

        if mode:
            # 注册本地数据源
            register_jdbc_default_datasource()
            # 注册kafka默认的数据源
            register_kafka_default_datasource()
            # 启动FD改变事件消息监听
            FactDatasourceChangeListener.instance().start()
            # 获取本地操作实例
            self.__instance = DistributedFactDatasourceProxy.instance()
        else:
            self.__instance = CentralizedFactDatasourceProxy.instance()
        return self.__instance


def fd_client_proxy() -> FactDatasourceProxy:
    return FactDatasourceProxyFactory.instance().get_fd_client_proxy()



# 定时清除日志文件的定时器
import logging
import os
import re
import shutil
from datetime import datetime

from apscheduler.schedulers.background import BackgroundScheduler

from config.annotation import Value
from schedule.utils.log import TaskLogFileAppender

scheduler = BackgroundScheduler()

logger = logging.getLogger('Ficus')

class TaskLogCleanScheduler:
    """
    定时清除日志文件的定时器
    """

    @Value("${log.retention.days:3}")
    def log_retention_days(self):
        pass

    @Value("${log.path:/sobeyHiveLogs/sdcLog/JOB/}")
    def base_log_path(self):
        pass

    def __init__(self) -> None:
        TaskLogFileAppender.init_log_path(self.base_log_path())


    def _run_(self):
        """
        清理日志文件
        :return:
        """
        if self.log_retention_days() < 1:
            return

        # 找到日志的目录
        logPath = TaskLogFileAppender.get_log_path()

        logger.info(f"过期日志清除器开始清理:{logPath} 下的超过{self.log_retention_days()}天的日志")

        today = datetime.now()

        try:
            with os.scandir(logPath) as it:
                for entry in it:
                    if re.fullmatch(r"\d{4}-\d{2}-\d{2}",entry.name) is not None:
                        # 找到文件夹了
                        fileName = entry.name
                        try:
                            logFileCreateDate = datetime.strptime(fileName,"%Y-%m-%d")
                        except:
                            # 文件名不合法
                            return
                        res = today - logFileCreateDate
                        if res.days>=self.log_retention_days():
                            # 超过了日期了,进行删除
                            shutil.rmtree(entry.path, ignore_errors=True)
                            logger.info(f"过期日志清除器,删除过期日志:{entry.path}")
        except Exception as e:
            logger.error(f"过期日志清除器,清理日志失败.{str(e)}")


    def start(self):
        """
        开启清理定时任务 每天晚上 2点30 执行清除.
        :return:
        """
        if scheduler.get_job("task_log_clean") is None:
            scheduler.add_job(self._run_, 'cron', minute='30',second='0',hour='2',id="task_log_clean")
            scheduler.start()
            return True

        return False



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/30
from unittest import TestCase

from libs.utils import Singleton, str_placeholder_replace


class TestSingletonClass(Singleton):
    def __init__(self):
        print('123+++')


class TestSingleton(TestCase):

    def test_instance(self):
        a = TestSingletonClass.instance()
        b = TestSingletonClass.instance()
        print(f'是否是同一个对象{a == b}')


class TestStrPlaceholderReplace(TestCase):

    def test_str_placeholder_replace(self):
        # value = '${${b}}'
        # value = 'a${b}c${d}_${d${e}}_${${f}}abcd'
        value = '''
        }${a}{
        'a': 'av',
         'b': '${b}',
         'c': [1,2,${f}]
         }}}'''
        value_map = {
            'b': 'b',
            'd': 'd',
            'e': 'e',
            'de': '#',
            'f': 'de'
        }
        result = str_placeholder_replace('${', '}', value, value_map)
        print(result)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import logging

from munch import Munch

from api.model.Page import Page
from factdatasource.dao.FactDatasource import customer_dao_context_holder as costomer_context
from factdatasource.dao.es.EsDatasourceDao import EsDatasourceDao
from factdatasource.execptions import FDExecuteException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext
from factdatasource.persistence.SqlableDeleteCondition import SqlableDeleteCondition

log = logging.getLogger('Ficus')


class EsFactDatasourceContext(AbstractFactDatasourceContext, SqlableDeleteCondition):
    DEFAULT_INDEX_TYPE = 'SobeyCube'

    @property
    def dao(self) -> EsDatasourceDao:
        return EsDatasourceDao.instance()

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            result = self.dao.count(index=target)
            size = 0 if result is None else result
            return size
        except Exception as e:
            error = f'事实库{source_name}执行size操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            result = self.dao.count(index=target)
            size = 0 if result is None else result
            return size <= 0
        except Exception as e:
            error = f'事实库{source_name}执行is_empty操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)
            size = size if size else 0
            result = self.dao.simple_query(index=target, size=size)
            return result
        except Exception as e:
            error = f'事实库{source_name}执行collect操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        query = self.search_condition_sql(target, condition_groups)
        log.debug(f'事实库{source_name}执行delete操作SQL：{query}')
        # 这里需要把sql转换成为es的json语句
        # TODO 暂时先不支持  这里如果找不到开源的SQL转换工具，就只能自己写一个了
        raise FDExecuteException(f'事实库{source_name}执行query操作查询语句错误，SQL：{query}')

    def query(self, query: str, parameters: dict = None):
        """
        使用查询语句查询数据
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target

        page = None
        if parameters and 'pageNum_' in parameters.keys() and 'pageSize_' in parameters.keys():
            # 需要分页的
            param = {
                'pageNum': parameters.get('pageNum_'),
                'pageSize': parameters.get('pageSize_'),
                'needCount': parameters.get('needCount_') if parameters.get('needCount_') else False
            }
            page = Page(**param)

        # 参数占位符处理
        query = self._fd_placeholder_replace(query, parameters)

        # 处理SQL空格
        query = query.strip()

        if query.startswith('{') and query.endswith('}'):
            # 说明是json方式
            # 处理下分页
            if page:
                if 'from' not in query:
                    query['from'] = page.start_row
                if 'size' not in query:
                    query['size'] = page.pageSize
        else:
            # TODO 暂时先不支持  这里如果找不到开源的SQL转换工具，就只能自己写一个了
            raise FDExecuteException(f'事实库{source_name}执行query操作查询语句错误，SQL：{query}')

        # 进行查询
        try:
            log.debug(f'事实库{source_name}执行query操作SQL：{query}')
            costomer_context.set_source(source_name)
            query_result = self.dao.query(index=target, body=query)
        except Exception as e:
            error = f'事实库{source_name}执行query操作SQL：{query} 发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

        # 处理查询结果
        result = []
        if query_result and query_result['hits']:
            hits = query_result['hits']
            # 取结果
            if hits['hits']:
                result = [source['_source'] for source in hits['hits']]
            # 取总数
            total = None
            if page:
                if page.need_count():
                    total = hits['total']
                page.set_result(result, total)
                return page
        return Page.no_page(result)

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr=[]
        try:
            last_index = len(result_list) - 1
            costomer_context.set_source(source_name)
            primary_keys = self.get_primary_keys()
            for index, result in enumerate(result_list):
                # 寻找id字段
                if not isinstance(result, dict):
                    result = vars(result)

                primary_values_key = self.get_primary_values_key(primary_keys, result)
                if not primary_values_key:
                    # 没设置id,只能用upsert,create要求id必填
                    self.dao.upsert(index=table, doc_type=self.DEFAULT_INDEX_TYPE, body=result, refresh='true')
                    rrr.append(Munch({"success": True}))
                    continue

                # 最后一个再刷新
                if index < last_index:
                    self.dao.create(index=table, doc_type=self.DEFAULT_INDEX_TYPE, id=primary_values_key, body=result,
                                    refresh='wait_for')
                else:
                    self.dao.create(index=table, doc_type=self.DEFAULT_INDEX_TYPE, id=primary_values_key, body=result,
                                    refresh='true')
                rrr.append(Munch({"success": True}))
        except Exception as e:
            error = f'事实库{source_name}执行inserts操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()
        return rrr

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr = []
        try:
            costomer_context.set_source(source_name)
            # TODO 暂时没有考虑事务
            primary_keys = self.get_primary_keys()
            for index, result in enumerate(result_list):
                # 寻找id字段
                if not isinstance(result, dict):
                    result = vars(result)

                primary_values_key = self.get_primary_values_key(primary_keys, result)
                if not primary_values_key:
                    # 没设置id,只能用upsert,create要求id必填
                    log.warning(f'事实库{source_name}执行updates操作时未配置id字段,放弃该数据的更新：{result}')
                    continue

                # 这里有可能最后一个没有设置主键，导致跳过最后一条数据的刷新，所以这里就每条都设置刷新
                self.dao.partial_update(index=table, doc_type=self.DEFAULT_INDEX_TYPE, id=primary_values_key,
                                        body=result, refresh='true')
                rrr.append(Munch({"success": True}))
        except Exception as e:
            error = f'事实库{source_name}执行updates操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()
        return rrr

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        source_name = self.fd().get_source_name()
        rrr = []
        try:
            last_index = len(result_list) - 1
            costomer_context.set_source(source_name)
            primary_keys = self.get_primary_keys()
            for index, result in enumerate(result_list):
                # 寻找id字段
                if not isinstance(result, dict):
                    result = vars(result)

                primary_values_key = self.get_primary_values_key(primary_keys, result)
                if not primary_values_key:
                    # 没设置id,只能用upsert,create要求id必填
                    log.warning(f'事实库{source_name}执行upserts操作时未配置主键字段：{result}。')

                # 最后一个再刷新
                if index < last_index:
                    self.dao.upsert(index=table, doc_type=self.DEFAULT_INDEX_TYPE, id=primary_values_key, body=result,
                                    refresh='wait_for')
                else:
                    self.dao.upsert(index=table, doc_type=self.DEFAULT_INDEX_TYPE, id=primary_values_key, body=result,
                                    refresh='true')
                rrr.append(Munch({"success":True}))
        except Exception as e:
            error = f'事实库{source_name}执行upserts操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()
        return rrr

    def delete_all(self):
        """
        清空数据
        :return:
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        try:
            costomer_context.set_source(source_name)

            body = {
                "query": {
                    "match_all": {}
                }
            }
            self.dao.delete_by_query(index=target, body=body)
        except Exception as e:
            error = f'事实库{source_name}执行delete_all操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是es的查询语句
        :return:
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        log.debug(f'事实库{source_name}执行delete操作SQL：{query}')
        try:
            costomer_context.set_source(source_name)
            self.dao.delete_by_query(index=target, body=query)
        except Exception as e:
            error = f'事实库{source_name}执行delete操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def delete_conditions(self, condition_groups: list):
        """
        根据删除条件,构造删除语句
        :param condition_groups:
        :return:
        """
        source_name = self.fd().get_source_name()
        target = self.fd().target
        query = self.delete_condition_sql(target, condition_groups)
        log.debug(f'事实库{source_name}执行delete操作SQL：{query}')
        # 这里需要把sql转换成为es的json语句
        # TODO 暂时先不支持  这里如果找不到开源的SQL转换工具，就只能自己写一个了
        raise FDExecuteException(f'事实库{source_name}执行query操作查询语句错误，SQL：{query}')
        # try:
        #     costomer_context.set_source(source_name)
        #     self.dao.delete_by_query(index=target, body=query, doc_type=self.DEFAULT_INDEX_TYPE, analyze_wildcard=True)
        # except Exception as e:
        #     error = f'事实库{source_name}执行delete操作发生异常, {str(e)}'
        #     log.error(error)
        #     raise FDExecuteException(error)
        # finally:
        #     costomer_context.clear_source()



from api.exceptions import IllegalArgumentException, ServiceInnerException
from api.model import OutputWrapper
from factdatasource import FactDatasourceProxyService
from schedule.utils.log import FrameworkHandlerLogger


class IBaseOutputer:

    def __logResult(self,task_logger,results):
        if results is not None and len(results) > 0:
            results = filter(lambda x: not x.success, results)
            for res in results:
                task_logger.log(f"{res.error} data:{res.content}")

    def put_in_cache(self, code, insert_cache: dict, update_cache, upsert_cache, output_fd, poll: OutputWrapper):
        """
        把数据放入缓存中
        :param code:
        :param insert_cache:
        :param update_cache:
        :param upsert_cache:
        :param output_fd:
        :param poll:
        :return:
        """
        from api.model.OutputWrapper import OperationEnum
        if poll.operation() == OperationEnum.INSERT:
            insert_cache[output_fd] = insert_cache.get(output_fd, [])
            insert_cache[output_fd].append(poll.content())

        elif poll.operation() == OperationEnum.UPDATE:
            update_cache[output_fd] = update_cache.get(output_fd, [])
            update_cache[output_fd].append(poll.content())

        elif poll.operation() == OperationEnum.UPSERT:
            upsert_cache[output_fd] = upsert_cache.get(output_fd, [])
            upsert_cache[output_fd].append(poll.content())

        elif poll.operation() == OperationEnum.CLEAN:
            try:
                FactDatasourceProxyService.fd_client_proxy().delete_all(output_fd)
            except Exception as e:
                raise ServiceInnerException(f"执行失败,Code: {code}的执行器,清空数据发生错误:{str(e)}")
        elif poll.operation() == OperationEnum.DELETE:
            try:
                FactDatasourceProxyService.fd_client_proxy().delete(output_fd, poll.content())
            except Exception as e:
                raise ServiceInnerException(f"执行失败,Code: {code}的执行器,删除数据发生错误:{str(e)}")
        elif poll.operation() == OperationEnum.FLUSH:
            self.flush_cache(code, insert_cache, update_cache, upsert_cache)

    def flush_cache(self, code, insert_cache, update_cache, upsert_cache):
        """
        把缓存中的数据 发送到FD中
        :param code:
        :param insert_cache:
        :param update_cache:
        :param upsert_cache:
        :return:
        """

        from schedule.utils.log.TaskLogger import TaskLogger
        from schedule.utils.log import TaskLogFileAppender
        task_logger = None
        if TaskLogFileAppender.get_log_file_name() is not None:
            task_logger = TaskLogger(TaskLogFileAppender.get_log_file_name())

        try:
            for key, value in insert_cache.items():
                if self.is_killed():
                    FrameworkHandlerLogger.log_warn("任务被强制停止,停止变更数据")
                    task_logger.log("任务被强制停止,停止变更数据")
                    return

                if len(value) > 0:
                    results = FactDatasourceProxyService.fd_client_proxy().inserts(key, value)
                    self.__logResult(task_logger, results)
                    value.clear()

            for key, value in update_cache.items():
                if self.is_killed():
                    FrameworkHandlerLogger.log_warn("任务被强制停止,停止变更数据")
                    task_logger.log("任务被强制停止,停止变更数据")
                    return

                if len(value) > 0:
                    results = FactDatasourceProxyService.fd_client_proxy().updates(key, value)
                    self.__logResult(task_logger, results)
                    value.clear()

            for key, value in upsert_cache.items():
                if self.is_killed():
                    FrameworkHandlerLogger.log_warn("任务被强制停止,停止变更数据")
                    task_logger.log("任务被强制停止,停止变更数据")
                    return

                if len(value) > 0:
                    results = FactDatasourceProxyService.fd_client_proxy().save_or_updates(key, value)
                    self.__logResult(task_logger, results)
                    value.clear()
        except Exception as e:
            import traceback
            raise ServiceInnerException(f"执行失败,Code: {code}的执行器,保存数据发生错误:\n{traceback.format_exc()}")

    def find_output_fd(self, output_fds: list, index: int):
        """
        从多个fd中,找到需要的那个fd
        :param output_fds:
        :param index:
        :return:
        """
        if len(output_fds) == 0:
            raise IllegalArgumentException("获取输出FD失败,输出为空")

        if len(output_fds) <= index:
            raise IllegalArgumentException(f"获取输出FD失败,可选输出个数:{len(output_fds)} 期望输出:{index}")

        return output_fds[index]

    def is_killed(self):
        """
        判断是否已经被杀掉了
        :return:
        """
        return False



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
from sqlalchemy.engine import ResultProxy, RowProxy
from sqlalchemy.orm import Session

from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import Datasource
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from libs.utils import Singleton


class JdbcDatasourceDao(Singleton):

    @property
    def datasource(self) -> Datasource:
        data_source = get_multiple_datesource(FactDatasourceTypeEnum.JDBC)
        return data_source

    @property
    def client(self) -> Session:
        return self.datasource.get_client()

    def select_one(self, sql: str, params: dict = None) -> dict:
        """
        查询一行结果
        :param sql:
        :param params:
        :return:
        """
        result = self.__execute(sql, params)
        row = result.fetchone()
        return self.__result_to_dict(row)

    def select_all(self, sql: str, params: dict = None) -> list:
        """
        查询所有结果
        :param sql:
        :param params:
        :return:
        """
        result = self.__execute(sql, params)
        rows = result.fetchall()
        return self.__result_to_list(rows)

    def select_total(self, sql: str, params: dict = None):
        """
        查询sql的返回结果有多少条
        :param sql:
        :param params:
        :return:
        """
        count_result = self.__execute(sql, params)
        count_result.close()
        # TODO 这个方式在sqlserver中没有作用,需要新的方法，暂时先这样
        # 详见 https://usyiyi.github.io/sqlalchemy-docs-zh/dialects/mssql.html#rowcount-support-orm-versioning
        return count_result.rowcount

    def select_num(self, sql: str, params: dict = None):
        """
        查询一行一个结果，比如 select count(*) from table  这种
        :param sql:
        :param params:
        :return: str or num or obj
        """
        result = self.__execute(sql, params)
        row = result.fetchone()
        return row.values()[0] if len(row.values()) > 0 else None

    def execute(self, sql: str, params: dict = None) -> ResultProxy:
        """
        外部直接执行sql
        :param sql:
        :param params:
        :return:
        """
        return self.__execute(sql, params)

    def __execute(self, sql: str, params: dict = None) -> ResultProxy:
        """
        统一执行sql
        :param sql:
        :param params:
        :return:
        """
        return self.client.execute(sql, params)

    def __result_to_list(self, rows: list):
        """
        将多行查询结果转换为 list<dict>
        :param result: list<dict>
        :return:
        """
        if not rows:
            return []
        return [dict(zip(row.keys(), row)) for row in rows]

    def __result_to_dict(self, row: RowProxy):
        """
        将单行查询结果转换为dict
        :param row:
        :return:
        """
        if not row:
            return dict()
        return dict(zip(row.keys(), row))




#-*-coding=utf-8-*-
from __future__ import absolute_import

# 中间件
#BROKER_URL = 'redis://localhost:6379/0'
# BROKER_URL = 'amqp://admin:admin@172.16.148.211:5672'

# 结果存储
#CELERY_RESULT_BACKEND = 'redis://127.0.0.1:6379/0'
# CELERY_RESULT_BACKEND = 'db+mysql://root:@localhost:3306/celery'

# 任务队列的名字
#CELERY_DEFAULT_QUEUE = '20013'

# 不提前缓存4个任务,收一个任务,做一个任务
CELERYD_PREFETCH_MULTIPLIER = 1

# CELERYD_CONCURRENCY = 1     # 并发任务数

# 记录任务的开始状态 否则就只有PENDING状态.   并且这个状态只任务结束24小时内有效.
CELERY_TRACK_STARTED = True

CELERY_TASK_SEND_SENT_EVENT = True

CELERY_SEND_EVENTS =True

# 忽略执行的结果(不能配置成True,否则获取不了状态和结果)
CELERY_IGNORE_RESULT = False

# 异步任务
CELERY_IMPORTS = (
    "cloudcelery.CeleryOnRequest"
)


from queue import Queue

from api.exceptions import ServiceInnerException
from api.model import OutputWrapper
from api.model.OtherTargetOutputPipe import OtherTargetOutputPipe
from api.model.ReadWriteLock import ReadWriteLock


class BatchOutputPipe:
    __output_stream: Queue = None
    __output_count = 0
    __lock = ReadWriteLock()

    def __init__(self, output_stream, output_count):
        self.__output_stream = output_stream
        self.__output_count = output_count

    def clean(self):
        self.__lock.acquire_read()
        try:
            self.__output_stream.put_nowait(OutputWrapper.CLEAN())
        except Exception as e:
            raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
        finally:
            self.__lock.release_read()

        return OtherTargetOutputPipe(self.__output_stream, 1, self.__output_count, self.__lock)

    def flush(self):
        self.__lock.acquire_read()
        try:
            self.__output_stream.put_nowait(OutputWrapper.FLUSH())
        except Exception as e:
            raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
        finally:
            self.__lock.release_read()

    def output_for_insert(self, serializable):
        self.__lock.acquire_read()
        try:
            self.__output_stream.put_nowait(OutputWrapper.INSERT(serializable))
        except Exception as e:
            raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
        finally:
            self.__lock.release_read()

        return OtherTargetOutputPipe(self.__output_stream, 1, self.__output_count, self.__lock)

    def output_for_delete(self, query):
        self.__lock.acquire_read()
        try:
            self.__output_stream.put_nowait(OutputWrapper.DELETE(query))
        except Exception as e:
            raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
        finally:
            self.__lock.release_read()

        return OtherTargetOutputPipe(self.__output_stream, 1, self.__output_count, self.__lock)

    def output_for_update(self, serializable):
        self.__lock.acquire_read()
        try:
            self.__output_stream.put_nowait(OutputWrapper.UPDATE(serializable))
        except Exception as e:
            raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
        finally:
            self.__lock.release_read()

        return OtherTargetOutputPipe(self.__output_stream, 1, self.__output_count, self.__lock)

    def output_for_upsert(self, serializable):
        self.__lock.acquire_read()
        try:
            self.__output_stream.put_nowait(OutputWrapper.UPSERT(serializable))
        except Exception as e:
            raise ServiceInnerException(f"放入输出队列失败:{str(e)}")
        finally:
            self.__lock.release_read()

        return OtherTargetOutputPipe(self.__output_stream, 1, self.__output_count, self.__lock)

    def skip_output(self):
        return OtherTargetOutputPipe(self.__output_stream, 1, self.__output_count, self.__lock)

    def is_all_empty(self):
        self.__lock.acquire_write()
        try:
            return self.__output_stream.empty()
        finally:
            self.__lock.release_write()



import warnings
from abc import abstractmethod

from api.handler.ce.ISimpleScriptCE import ISimpleScriptCE
from api.model.FdInputPipe import FdInputPipe

warnings.warn("'ISampleScriptCE'类已被废弃,请使用'ISimpleScriptCE'类代替 ", DeprecationWarning, 2)
class ISampleScriptCE(ISimpleScriptCE):
    """
    Python脚本式的CE的基类
    """

    @abstractmethod
    def do_compute(self,source_fds:FdInputPipe, params:dict):
        """
        计算器的业务逻辑
        :param source_fds: 输入源FD
        :param params: 计算器上下文参数
        :return: 返回一个 OutputWrapper的 数组
        """
        warnings.warn("'ISampleScriptCE'类已被废弃,请使用'ISimpleScriptCE'类代替 ", DeprecationWarning, 2)
        pass


from enum import Enum


class LogicalOperator(Enum):
    OR = "OR"
    AND = "AND"


class CalculationOperator(Enum):
    EQUAL = "="
    NOTEQUAL = "!="
    GT = ">"
    GTE = ">="
    LT = "<"
    LTE = "<="
    IN = "in"
    LIKE = "like"


class ConditionGroup:
    def __init__(self, operator: LogicalOperator, conditions: list):
        """

        :param operator:
        :param conditions:
        """
        self.conditions = conditions
        self.operator = operator


class Condition:

    def __init__(self, logical_operator: LogicalOperator, key: str, value, calculation_operator: CalculationOperator):
        """

        :param logical_operator:
        :param key:
        :param value:
        :param calculation_operator:
        """
        self.logical_operator: LogicalOperator = logical_operator
        self.key = key
        self.value = value
        self.calculation_operator: CalculationOperator = calculation_operator



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/3/11
import json

from sqlalchemy.engine import ResultProxy
from sqlalchemy.orm import Session

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasourceTypeEnum
from api.model.GraphNode import GraphNode
from api.model.GraphRelation import GraphRelation
from factdatasource.dao.FactDatasource import Datasource
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from libs.utils import Singleton, uuid4


class GraphDatasourceDao(Singleton):

    @property
    def datasource(self) -> Datasource:
        data_source = get_multiple_datesource(FactDatasourceTypeEnum.JDBC)
        return data_source

    @property
    def client(self) -> Session:
        return self.datasource.get_client()

    # region 创建
    def create_node(self, table: str, graph_node: GraphNode) -> str:
        """
        创建一个节点
        :param table:
        :param graph_node:
        :return: 节点id
        """
        if not table or not graph_node:
            raise IllegalArgumentException('graph创建节点table或graph_node不能为空')

        if not graph_node.guid:
            guid = uuid4()
            graph_node.guid(guid)
        node_param = self.__convert_node(graph_node)

        sql = f'INSERT NODE(ID, TYPE, RESOURCEID,GROUP,TAGS) INTO {table} VALUES(:id,:type, :resourceId, :group, :tags)'
        result: ResultProxy = self.client.execute(sql, node_param)
        if result.rowcount > 0:
            return graph_node.guid
        return None

    def create_nodes(self, table: str, graph_nodes: list):
        """
        创建多个节点
        :param table:
        :param graph_nodes:
        :return:
        """
        if not table or not graph_nodes:
            raise IllegalArgumentException('graph创建节点table或graph_nodes不能为空')
        result = []
        for graph_node in graph_nodes:
            node_id = self.create_node(table, graph_node)
            result.append(node_id)
        return result

    def create_relation(self, table: str, graph_relation: GraphRelation) -> str:
        """
        创建一个节点
        :param table:
        :param graph_relation:
        :return: 节点id
        """
        if not table or not graph_relation or not graph_relation.from_ or not graph_relation.to:
            raise IllegalArgumentException('graph创建关系参数错误')

        if not graph_relation.guid:
            guid = uuid4()
            graph_relation.guid(guid)
        relation_param = self.__convert_relation(graph_relation)

        sql = f'INSERT RELATION(ID, TYPE ,GROUP, TAGS) INTO {table} VIA NODE1(ID = :id1) TO NODE2(ID = :id2) ' \
              f'VALUES(:id,:type,:group,:tags)'
        result: ResultProxy = self.client.execute(sql, relation_param)
        if result.rowcount > 0:
            return graph_relation.guid
        return None

    def create_relations(self, table: str, graph_relations: list):
        """
        创建多个关系
        :param table:
        :param graph_relations:
        :return:
        """
        if not table or not graph_relations:
            raise IllegalArgumentException('graph创建关系参数错误')

        result = []
        for graph_relation in graph_relations:
            relation_id = self.create_relation(table, graph_relation)
            result.append(relation_id)
        return result

    # endregion

    # region Exists
    def node_exists(self, table: str, id: str) -> bool:
        """
        节点是否存在
        :param id:
        :return:
        """
        return self.get_one_node(table, id) != None

    # endregion

    # region get
    def get_one_node(self, table: str, id: str):
        """
        根据id获取一个节点
        :param table:
        :param id:
        :return:
        """
        param = {'ID': id}
        sql = f'SELECT * FROM NODE {table} WHERE ID = :ID'
        result: ResultProxy = self.client.execute(sql, param)
        row = result.fetchone()
        if not row:
            return None
        return dict(zip(row.keys(), row))

    def get_one_relation(self, table: str, id: str):
        """
        根据id获取一个关系
        :param table:
        :param id:
        :return:
        """
        param = {'ID': id}
        sql = f'SELECT * FROM RELATION {table} WHERE ID = :ID'
        result: ResultProxy = self.client.execute(sql, param)
        row = result.fetchone()
        if not row:
            return None
        return dict(zip(row.keys(), row))

    # endregion

    # region delete
    def delte_node_by_id(self, table: str, id: str):
        """
        根据id删除一个节点
        :param table:
        :param id:
        :return:
        """
        param = {'ID': id}
        sql = f'DELETE NODE FROM {table} WHERE ID = :ID'
        result: ResultProxy = self.client.execute(sql, param)
        return result.rowcount > 0

    def delte_relation_by_id(self, table: str, id: str):
        """
        删除关系
        :param table:
        :param id:
        :return:
        """
        param = {'ID': id}
        sql = f'DELETE RELATION FROM {table} WHERE ID = :ID'
        result: ResultProxy = self.client.execute(sql, param)
        return result.rowcount > 0

    def delete_by_query(self, query: str, param: dict = None):
        """
        根据查询来删除
        :param query:
        :param param:
        :return:
        """
        result: ResultProxy = self.client.execute(query, param)
        return result.rowcount > 0
    # endregion

    # region update
    def update_relation_tags(self, table: str, group_guid: str, graph_relation: GraphRelation):
        """
        修改关系标签
        :param table:
        :param group_guid:
        :param graph_relation:
        :return:
        """
        param = {
            'group': group_guid,
            'id': graph_relation.guid,
            'tags': json.dumps(graph_relation.tags) if graph_relation.tags else None,
        }
        sql = f'UPDATE RELATION FROM {table}  SET TAGS= :tags WHERE ID = :id AND GROUP = :group '
        result: ResultProxy = self.client.execute(sql, param)
        return result.rowcount > 0

    def update_node_tags(self, table: str, group_guid: str, graph_node: GraphNode):
        """
        修改关系标签
        :param table:
        :param group_guid:
        :param graph_node:
        :return:
        """
        param = {
            'group': group_guid,
            'id': graph_node.guid,
            'tags': json.dumps(graph_node.tags) if graph_node.tags else None,
        }
        sql = f'UPDATE NODE FROM {table}  SET TAGS= :tags WHERE ID = :id AND GROUP = :group '
        result: ResultProxy = self.client.execute(sql, param)
        return result.rowcount > 0
    # endregion

    # region query
    def query(self, sql: str, params: dict = None) -> list:
        """
        查询所有结果
        :param sql:
        :param params:
        :return:
        """
        result = self.client.execute(sql, params)
        rows = result.fetchall()
        if not rows:
            return []
        return [dict(zip(row.keys(), row)) for row in rows]

    def query_total(self, sql: str, params: dict = None):
        """
        查询sql的返回结果有多少条
        :param sql:
        :param params:
        :return:
        """
        count_result = self.client.execute(sql, params)
        count_result.close()
        return count_result.rowcount
    # endregion

    def __convert_node(self, graph_node: GraphNode):
        node = {
            'id': graph_node.guid,
            'group': graph_node.get('GROUP'),
            'type': graph_node.type,
            'resourceId': graph_node.guid,
            'tags': json.dumps(graph_node.tags) if graph_node.tags else None,
        }
        return node

    def __convert_relation(self, graph_relation: GraphRelation):
        relation = {
            'id': graph_relation.guid,
            'id1': graph_relation.from_.guid,
            'id2': graph_relation.to.guid,
            'group': graph_relation.get('GROUP'),
            'type': graph_relation.type,
            'tags': json.dumps(graph_relation.tags) if graph_relation.tags else None,
        }
        return relation



from api.model.MetaModelField import MetaModelField
from factdatasource.transformer.ITransformer import ITransformer

PREFIX = "__prefix__"
SUFFIX = "__suffix__"


class CompleteStringTransformer(ITransformer):

    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField):
        # 如果没有这个字段就不管他
        if (data is None or len(data) == 0) or target_model_field.fieldName not in data:
            return {target_model_field.fieldName: None}

        # 能进来的只能是字符串
        old_value = data[target_model_field.fieldName]

        if old_value is None:
            return {target_model_field.fieldName: None}

        prefix = params[PREFIX] if PREFIX in params else ""
        suffix = params[SUFFIX] if SUFFIX in params else ""

        return {target_model_field.fieldName: f"{prefix}{old_value}{suffix}"}



from api.model.MetaModelField import MetaModelField
from factdatasource.transformer.ITransformer import ITransformer

LETTER_CASE_KEY = "__case__"


class NumberToChineseTransformer(ITransformer):
    """
    数字转中文的处理器
    """

    def transform_field(self, data: dict, params: dict, target_model_field: MetaModelField):

        if (data is None or len(data) == 0) or target_model_field.fieldName not in data:
            return {target_model_field.fieldName: None}

        # 能进来的只能是字符串
        old_value = data[target_model_field.fieldName]

        if old_value is None:
            return {target_model_field.fieldName: None}

        # 获取大小写
        letter_case = params[LETTER_CASE_KEY] if LETTER_CASE_KEY in params else "lower"

        import cn2an
        return {target_model_field.fieldName: cn2an.an2cn(old_value, letter_case == "upper")}



import logging
import threading
import time
from collections import deque
from datetime import datetime
from queue import Queue, Empty

import client
from api.exceptions import IllegalArgumentException
from api.model.ResultVO import *
from schedule.utils.log import TaskLogFileAppender, FrameworkHandlerLogger

log = logging.getLogger('Ficus')


class BlockingQueue(Queue):
    """
    允许删除的阻塞队列
    """

    def _init(self, maxsize):
        self.queue = deque()

    def _qsize(self):
        return len(self.queue)

    def _put(self, item):
        self.queue.append(item)

    def _get(self):
        return self.queue.popleft()

    def remove(self, item):
        self.queue.remove(item)


class TaskThread(threading.Thread):
    """
    真正开始执行任务的线程
    """

    def __init__(self, handler, actor_port):
        """
        执行器线程的构造函数
        :param handler: ITaskHandler
        :param actor_port:
        """
        threading.Thread.__init__(self, name="任务线程")

        if handler is None:
            raise IllegalArgumentException("创建TaskThread失败,传入的ITaskHandler为空,无法执行")
        self.handler = handler
        self.actor_port = actor_port
        self.to_stop = False
        self.stop_reason = None
        self.running = False
        self.trigger_queue: BlockingQueue = BlockingQueue()  # 使用了Python中的队列
        self.trigger_log_id_set = set()  # TODO ConcurrentHashSet
        self.trigger_log_id_dict = {}

    def get_handler(self):
        """
        获取handler
        :return:
        """
        return self.handler

    def push_trigger_queue(self, trigger_param):
        """
        把任务放入 任务队列中去
        :param trigger_param:
        :return:
        """
        if trigger_param.logId in self.trigger_log_id_set:
            return ResultVO(code=FAIL_CODE, msg=f"任务已重复,LogId:{trigger_param.log_id}")

        # 放入队列中
        self.trigger_log_id_set.add(trigger_param.logId)
        self.trigger_queue.put_nowait(trigger_param)
        import config
        return ResultVO(code=SUCCESS_CODE,
                        msg=f"success,执行器ip:{config.server_ip or config.find_host_ip()}:{config.server_port or 5000}",
                        content="success")

    def stop(self, stop_reason):
        """
        强制杀掉任务
        :param stop_reason:
        :return:
        """
        log.info(f"任务TaskThread终止,原因:{stop_reason}")
        self.to_stop = True
        self.stop_reason = stop_reason

        if self.handler is not None:
            try:
                self.handler.kill()
                self.handler.task_logger = None
            except Exception as e:
                FrameworkHandlerLogger.log_warn(f"kill Handler 失败,", e)

    def remove_trigger_queue(self, log_id):
        """
        从等待队列中删除任务
        :param log_id:
        :return:
        """
        if log_id in self.trigger_log_id_set:
            # 说明任务确实在这里面
            for next in list(self.trigger_queue.queue):
                if log_id == next.logId:
                    # 说明找到任务了
                    self.trigger_queue.remove(next)
                    self.trigger_log_id_set.remove(log_id)
                    # 需要修改任务的状态,
                    self.__update_task_status_to_finished(log_id, ResultVO(FAIL_CODE, "任务在等待队列被手动强制停止"), True, 1)
                    return True
        else:
            # 2019-12-04 14:31:08 孙翔 数据库记录是在这台机器中执行的,但结果在triggerLogIdSet里面有没有.
            # 说明 这台机器有可能是重启过. 那么就直接更新状态
            # 需要修改任务的状态
            self.__update_task_status_to_finished(log_id, ResultVO(FAIL_CODE, "任务不在等待队列,可能执行器被重启.手动强制停止"), True, 1)
            return True
        return False

    def kill_doing_task(self, log_id):
        """
        强制杀掉handler,也就是当前的任务
        :return:
        """
        FrameworkHandlerLogger.log_info(f"任务TaskThread 强制杀掉当前执行的任务")
        self.to_stop = True
        self.stop_reason = "人工手动终止"
        # 停的时候也需要把handler里面的东西清理了
        if self.handler is not None:
            try:
                if (self.trigger_log_id_dict is not None) and (self.trigger_log_id_dict.get(log_id) is self.handler):
                    self.handler.kill()
                else:
                    self.__update_task_status_to_finished(log_id, ResultVO(FAIL_CODE, "任务不在当前执行队列中,可能执行器被重启.手动强制停止"),
                                                          True, 1)
                self.handler.task_logger = None
            except Exception as e:
                FrameworkHandlerLogger.log_warn(f"kill Handler 失败,", e)

    def is_running_or_has_queue(self):
        """
        判断任务线程的情况
        :return:
        """
        return self.running or not self.trigger_queue.empty()

    def handle_execute_result(self, task_param, execute_result: ResultVO):
        if execute_result.code == FAIL_CODE and task_param.retryTimes is not None and task_param.retryTimes > 0:
            # region 如果是失败的,并且有重试次数的,那么就需要做重试的判断
            if task_param.retryLogId is None:
                # 说明是第一次重试
                currentRetryTimes = 0
            else:
                # 计算失败的任务次数,就知道重试了几次了
                currentRetryTimes = client.ScheduleJobTaskLogClient.count_failed_task_log_by_retry_log_id(
                    task_param.retryLogId)

            if currentRetryTimes < task_param.retryTimes:
                # 执行完成,回调写入 执行结果
                self.__update_task_status_to_finished(task_param.logId, execute_result, False, 1)

                # 小于设定的重试次数,那么就需要重试
                # 重试的操作就是类似于手动触发一次调度

                # 把原来的参数还原回去
                param = task_param.actorParams
                param["retryLogId"] = task_param.retryLogId or task_param.logId
                if -1 != task_param.shardIndex and -1 != task_param.shardTotal:
                    param["shardingParam"] = f"{task_param.shardIndex}/{task_param.shardTotal}"

                if client.JobScheduleClient.trigger_job(task_param.jobId, param):
                    # 重新触发成功
                    FrameworkHandlerLogger.log_info(
                        f"处理jobId:{task_param.jobId}的任务实例:{task_param.retryLogId}  成功触发重试,当前重试次数:{currentRetryTimes}/{task_param.retryTimes}")
                else:
                    FrameworkHandlerLogger.log_error(
                        f"处理jobId:{task_param.jobId}的任务实例:{task_param.retryLogId}  触发重试失败,当前重试次数:{currentRetryTimes}/{task_param.retryTimes}")
            else:
                # 执行完成,回调写入 执行结果
                self.__update_task_status_to_finished(task_param.logId, execute_result, True, 1)
                # 等于大于了重试次数,那么就不重试了.
                FrameworkHandlerLogger.log_warn(
                    f"处理jobId:{task_param.jobId}的任务实例:{task_param.retryLogId}  已超过最大重试次数:{task_param.retryTimes} 不继续进行重试")
            # endregion
        else:
            # 执行完成,回调写入 执行结果
            self.__update_task_status_to_finished(task_param.logId, execute_result, True, 1)

        # 这里判断执行次数,如果判断到已经到达执行次数的上限,就要把这个cron调度给停止了
        if task_param.limitTimes is not None and task_param.limitTimes > 0:
            # 说明这个任务是只执行有限次数的任务. 那么就需要查询这个Job已经执行成功了的任务的次数
            success_times = client.ScheduleJobTaskLogClient.count_success_task_log_by_job(task_param.jobId,
                                                                                          task_param.updateTime)
            if success_times >= task_param.limitTimes:
                # 说明已经到达了执行次数的限制.那么就需要停止这个调度
                if client.JobScheduleClient.stop(task_param.jobId):
                    self.stop("达到任务执行次数上限,完成job")

    def run_sync(self, task_param):
        """
        同步跑任务
        :param task_param:
        :return:
        """
        try:
            self.running = True
            return self.__inner_run(task_param)
        finally:
            self.running = False

    def run(self):
        """
        周期执行的东西
        :return:
        """

        log_id = None
        while not self.to_stop:
            self.running = False
            # 3秒等待一个任务
            task_param = None
            try:
                task_param = self.trigger_queue.get(timeout=3)
            except Empty:
                pass

            if task_param is None:
                # 没有任务
                continue

            # 有任务, 修改状态
            self.running = True
            log_id = task_param.logId
            self.trigger_log_id_set.remove(log_id)
            self.__inner_run(task_param)

        while self.trigger_queue is not None and not self.trigger_queue.empty():
            try:
                task_param = self.trigger_queue.get_nowait()
            except Empty:
                return
            if task_param is not None:
                self.__update_task_status_to_finished(task_param.logId,
                                                      ResultVO(FAIL_CODE,
                                                               f"{self.stop_reason}  处理logId:{log_id} 任务尚未执行,在调度队列中被终止:"),
                                                      True,
                                                      1)

    def count_trigger_queue(self) -> int:
        """
        获取等待队列的长度
        :return:
        """
        return len(self.trigger_queue.queue)

    # region 私有方法

    def __inner_run(self, task_param):
        log_id = None
        execute_result = None
        try:
            if task_param is None:
                # 没有任务
                return

            log_id = task_param.logId

            FrameworkHandlerLogger.log_info(f"开始处理logId:{log_id}的任务")

            if task_param.actorParams is None:
                task_param.actorParams = {}

            task_param.actorParams["__logId__"] = log_id
            task_param.actorParams["__jobId__"] = task_param.jobId
            task_param.actorParams["__triggerTime__"] = task_param.triggerTime
            task_param.actorParams["__processLogId__"] = task_param.processLogId

            # 更新任务状态到正在执行
            import config
            self.__update_task_status_to_execute(log_id,
                                                 f"{config.server_ip or config.find_host_ip()}:{config.server_port or 5000}",
                                                 1)

            from schedule import ShardContext
            try:
                from schedule.ShardContext import Sharding
                ShardContext.set_sharding(Sharding(task_param.shardIndex, task_param.shardTotal))
                TaskLogFileAppender.prepare_to_log(datetime.strptime(task_param.triggerTime, "%Y-%m-%d %H:%M:%S"),
                                                   task_param.logId)
                from schedule.utils.log.TaskLogger import TaskLogger
                self.handler.task_logger = TaskLogger(TaskLogFileAppender.get_log_file_path(
                    datetime.strptime(task_param.triggerTime, "%Y-%m-%d %H:%M:%S"), task_param.logId))
                self.trigger_log_id_dict[task_param.logId] = self.handler
                execute_result = self.handler.execute(task_param.actorParams)
                if execute_result is None:
                    execute_result = FAIL
            except Exception as e:
                FrameworkHandlerLogger.log_error(f"处理logId:{log_id}出现错误:", e)
                if self.to_stop:
                    FrameworkHandlerLogger.log_error(f"任务线程强制停止:", e)
                    try:
                        self.handler.kill()
                    except:
                        pass
                execute_result = ResultVO(FAIL_CODE, str(e))
            finally:
                ShardContext.reset()
                TaskLogFileAppender.end_log()
                self.handler.task_logger = None
                self.trigger_log_id_dict.pop(task_param.logId)

            if not self.to_stop:
                self.handle_execute_result(task_param, execute_result)
            else:
                # 执行强制停止,回调写入 失败信息
                self.__update_task_status_to_finished(log_id,
                                                      ResultVO(FAIL_CODE, f"{self.stop_reason} 业务运行中,被强制终止"), True,
                                                      1)
        except Exception as e:
            if log_id is not None:
                FrameworkHandlerLogger.log_error(f"处理logId:{log_id}出现严重错误", e)
                self.__update_task_status_to_finished(log_id,
                                                      ResultVO(FAIL_CODE,
                                                               f"{self.stop_reason}  处理logId:{log_id} 出现严重错误:"),
                                                      True,
                                                      1)
        return execute_result

    def __update_task_status_to_execute(self, log_id, instance_address, deep):
        """
        更新任务的状态到正在执行中
        :param log_id:
        :param instance_address:
        :param deep:
        :return:
        """
        if deep > 3:
            # 重试3次还不行, 直接退出
            FrameworkHandlerLogger.log_error(f"更新logId:{log_id} 的状态,重试3次仍然无法连接到ficus,无法更新任务状态到正在执行.")
            return False

        try:
            # 把任务状态设置为执行中
            client.ScheduleJobTaskLogClient.update_task_status_to_execute(log_id, instance_address)
        except Exception:
            FrameworkHandlerLogger.log_error(f"更新logId:{log_id} 的状态,连接到ficus,无法更新任务状态到正在执行.次数:{deep}")
            time.sleep(3)
            deep += 1
            return self.__update_task_status_to_execute(log_id, instance_address, deep)

    def __update_task_status_to_finished(self, log_id, execute_result, triggerProcess, deep):
        """
        更新任务的状态到结束
        :param log_id:
        :param execute_result:
        :param deep:
        :return:
        """
        if deep > 3:
            # 重试3次还不行, 直接退出
            FrameworkHandlerLogger.log_error(f"更新logId:{log_id} 的状态,重试3次仍然无法连接到ficus,无法更新任务状态到完成.")
            return False

        try:
            # 把任务状态设置为执行中
            client.ScheduleJobTaskLogClient.update_task_status_to_finished(log_id, execute_result.to_dict(),
                                                                           triggerProcess)
        except Exception:
            FrameworkHandlerLogger.log_error(f"更新logId:{log_id} 的状态,连接到ficus,无法更新任务状态到到完成.次数:{deep}")
            time.sleep(3)
            deep += 1
            return self.__update_task_status_to_finished(log_id, execute_result, triggerProcess, deep)
    # endregion



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/31

from pymongo.collection import Collection
from pymongo.cursor import Cursor

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.dao.FactDatasource import Datasource
from factdatasource.dao.MultipleDatasourceHolder import get_multiple_datesource
from libs.utils import Singleton


class MongoDatasourceDao(Singleton):

    @property
    def datasource(self) -> Datasource:
        data_source = get_multiple_datesource(FactDatasourceTypeEnum.MONGO)
        return data_source

    def client(self, target) -> Collection:
        if not target:
            raise IllegalArgumentException('获取MONGO操作客户端参数不合法，target不能为空。')
        return self.datasource.get_client()[target]

    def count_documents(self, target, filter={}, **kwargs):
        """
        统计一个Collection中的条数
        :param target:
        :param filter:
        :param kwargs:
        :return:
        """
        count = self.client(target).count_documents(filter, **kwargs)
        return count

    def find(self, target, *args, **kwargs) -> list:
        """
        简单查询
        :param target:
        :param args:
        :param kwargs:
        :return:
        """
        client: Collection = self.client(target)
        result: Cursor = client.find(*args, **kwargs)
        return list(result)

    def map_reduce(self, target, map, reduce, out, **kwargs):
        """
        map_reduce查询
        :param target:
        :param map:
        :param reduce:
        :param out:
        :param kwargs:
        :return:
        """
        client: Collection = self.client(target)
        result = client.map_reduce(map, reduce, out, **kwargs)
        return result

    def aggregate(self, target, pipeline, **kwargs):
        """
        聚合管道查询
        :param target:
        :param pipeline:
        :param kwargs:
        :return: 这里返回的是查询到的所有结果
        """
        client: Collection = self.client(target)
        result: Cursor = client.aggregate(pipeline, **kwargs)
        return list(result)

    def delete_many(self, target, filter):
        """
        根据条件删除数据
        :param target:
        :param filter: mongo查询条件
        :return:
        """
        client: Collection = self.client(target)
        client.delete_many(filter)

    def drop(self, target):
        """
        删除collection
        :param target:
        :return:
        """
        client: Collection = self.client(target)
        client.drop()

    def insert(self, target, documents: list):
        """
        添加数据
        :param target:
        :param documents:
        :return:
        """
        if not documents:
            return
        client: Collection = self.client(target)
        client.insert_many(documents)

    def update_one(self, target, filter, document: dict, upsert=False):
        """
        修改数据
        :param target:
        :param filter:
        :param document:
        :param upsert: 数据不存在是否进行插入操作
        :return:
        """
        if not filter or not document:
            raise IllegalArgumentException('修改MONGO数据时参数不合法，filter、document不能为空。')
        client: Collection = self.client(target)
        update_doc = {"$set": document}
        client.update_one(filter, update_doc, upsert=upsert)






#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/21
import logging

from factdatasource.dao.FactDatasource import customer_dao_context_holder as costomer_context
from factdatasource.dao.kafka.KafkaDatasourceDao import KafkaDatasourceDao
from factdatasource.execptions import NotSupportedFDException, FDExecuteException
from factdatasource.persistence.AbstractFactDatasourceContext import AbstractFactDatasourceContext

log = logging.getLogger('Ficus')


class KafkaFactDatasourceContext(AbstractFactDatasourceContext):

    @property
    def dao(self) -> KafkaDatasourceDao:
        return KafkaDatasourceDao.instance()

    def size(self):
        """
        返回数据总长度
        :return: 数据条数:long
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def is_empty(self):
        """
        返回是否存在数据
        :return: boolean
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def collect(self, size: int):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :return: list
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def collect_conditions(self, size: int, condition_groups: list):
        """
        返回指定条数的数据
        :param size: 返回的条数
        :param condition_groups: 查询条件
        :return: list
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def query(self, query: str, parameters: dict = None):
        """
        使用查询语句查询数据
        :param query: 查询语句
        :param parameters: 查询参数
        :return: Page
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def inserts(self, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        if not result_list:
            return
        if not isinstance(result_list, list):
            result_list = [result_list]

        source_name = self.fd().get_source_name()
        topic = self.fd().target
        try:
            costomer_context.set_source(source_name)
            for result in result_list:
                self.dao.producer.produce(topic, result)
            self.dao.producer.flush(30)
        except Exception as e:
            error = f'事实库{source_name}执行size操作发生异常, {str(e)}'
            log.error(error)
            raise FDExecuteException(error)
        finally:
            costomer_context.clear_source()

    def updates(self, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def inserts_or_updates(self, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        self.inserts(result_list)

    def _single_thread_inserts(self, table: str, result_list: list):
        """
        批量保存数据,要求list里面的字段和数据库里面的字段一一对应
        :param result_list: 要保存的数据
        :return:
        """
        # 消息可能是有顺序的,因此先不并发
        raise NotSupportedFDException('kafka不支持该操作')

    def _single_thread_updates(self, table: str, result_list: list):
        """
        批量更新数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要修改的数据
        :return:
        """
        # 消息可能是有顺序的,因此先不并发
        raise NotSupportedFDException('kafka不支持该操作')

    def _single_thread_inserts_or_updates(self, table: str, result_list: list):
        """
        批量saveOrUpdate数据,,数据,要求list里面的字段和数据库里面的字段一一对应
        采用ByPrimaryKeySelective的方式,也就是主键必填,其他的字段非空就是要修改的
        :param result_list: 要添加或者需要修改的数据
        :return:
        """
        # 消息可能是有顺序的,因此先不并发
        raise NotSupportedFDException('kafka不支持该操作')

    def delete_all(self):
        """
        清空数据
        :return:
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def delete(self, query: str):
        """
        根据删除语句删除数据,query是完整的删除语句
        :return:
        """
        raise NotSupportedFDException('kafka不支持该操作')

    def delete_conditions(self, condition_groups: list):
        raise NotSupportedFDException('kafka不支持该操作')




# 日志获取的远程接口
from datetime import datetime

from schedule.utils.log import TaskLogFileAppender
from . import remote
from flask import request, jsonify

@remote.route('/remote/log-read-service/<string:log_id>/<int:trigger_time>', methods=['GET'])
def log_detail(trigger_time, log_id):
    from_line_num = int(request.args.get("fromLineNum", default=0))
    to_line_num = int(request.args.get("toLineNum", default=-1))
    instance_address = request.args.get("instanceAddress",default=None)
    # 获取日志文件
    return jsonify(TaskLogFileAppender.read_log(instance_address,datetime.fromtimestamp(trigger_time/1000),log_id,from_line_num,to_line_num))

@remote.route('/remote/log-read-service/tailable', methods=['GET'])
def tailable():
    """
    python的因为开启websocket或者sse非常的困难,因此暂时不支持实时的日志输出
    :return:
    """
    return jsonify(False)


from abc import abstractmethod

class ICacheAble:

    @abstractmethod
    def set_cache_value(self, key, value):
        """

        :param key:
        :param value:
        :return:
        """

    @abstractmethod
    def set_cache_value_if_absent(self, key, value):
        """

        :param key:
        :param value:
        :return:
        """

    @abstractmethod
    def get_cache_value(self, key):
        """

        :param key:
        :return:
        """

    @abstractmethod
    def delete_cache_value(self, key):
        """
        
        :param key:
        :return:
        """

    @abstractmethod
    def set_cache_value_from_process(self, key, value):
        """

        :param key:
        :param value:
        :return:
        """

    @abstractmethod
    def set_cache_value_if_absent_from_process(self, key, value):
        """

        :param key:
        :param value:
        :return:
        """

    @abstractmethod
    def get_cache_value_from_process(self, key):
        """

        :param key:
        :return:
        """

    @abstractmethod
    def delete_cache_value_from_process(self, key):
        """

        :param key:
        :return:
        """


import json
import logging

from api.model import LogMessage, ClassInfo
from client import HandlerLogClient
from config.annotation import Value

localLogger = logging.getLogger('Ficus')

@Value("${log.type:local}")        # 默认是本地日志
def log_type():
    pass

@Value("${actor.name:unknown}")        # 找到actorName
def actor_name():
    pass


def log_debug(message:str):
    """
    记录debug日志
    :param message:
    :return:
    """
    log_message = __generateLogMessage__(message)

    if "remote" == log_type():
        try:
            HandlerLogClient.log_debug(log_message)
        except:
            localLogger.error("记录远程日志失败,改为本地日志记录")
            localLogger.debug(json.dumps(log_message))
    else:
        localLogger.debug(message)

def log_info(message:str):
    """
    记录info日志
    :param message:
    :return:
    """
    log_message = __generateLogMessage__(message)

    if "remote" == log_type():
        try:
            HandlerLogClient.log_info(log_message)
        except:
            localLogger.error("记录远程日志失败,改为本地日志记录")
            localLogger.info(json.dumps(log_message))
    else:
        localLogger.info(message)

def log_warn(message:str,exception:Exception=None):
    """
    记录warn日志
    :param message:
    :return:
    """
    log_message = __generateLogMessage__(message,exception)

    if "remote" == log_type():
        try:
            HandlerLogClient.log_warn(log_message)
        except:
            localLogger.error("记录远程日志失败,改为本地日志记录")
            localLogger.warning(json.dumps(log_message))
    else:
        localLogger.warning(message if exception is None else f"{message},{str(exception)}")

def log_error(message:str,exception:Exception=None):
    """
    记录error日志
    :param message:
    :param exception:
    :return:
    """
    log_message = __generateLogMessage__(message,exception)

    if "remote" == log_type():
        try:
            HandlerLogClient.log_error(log_message)
        except:
            localLogger.error("记录远程日志失败,改为本地日志记录")
            localLogger.error(json.dumps(log_message))
    else:
        localLogger.error(message if exception is None else f"{message},{str(exception)}")


def __findClassInfo__(limit=1):
    import traceback
    stackTuple = traceback.extract_stack(limit=limit + 1)[0]
    class_info = ClassInfo()

    class_info.className = stackTuple[0]
    class_info.methodName = stackTuple[2]
    class_info.lineNumber = stackTuple[1]

    return class_info

def __generateLogMessage__(message:str,exception:Exception=None):
    log_message = LogMessage()
    log_message.actorName = actor_name()
    log_message.exception = exception
    log_message.message = message
    import threading
    log_message.threadName = threading.currentThread().getName()
    log_message.classInfo = __findClassInfo__(2)



import threading
from abc import abstractmethod
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from queue import Queue

import requests

from api.handler.IAsyncAble import IAsyncAble
from api.handler.ICacheAbleHandler import ICacheAbleHandler
from api.handler.ITaskHandler import ITaskHandler
from api.handler.outputer.IBatchOutputer import IBatchOutputer
from api.model.AsyncServiceRequest import AsyncServiceRequest, SendType, AcceptType
from api.model.BatchOutputPipe import BatchOutputPipe
from api.model.FdInputPipe import FdInputPipe
from api.model.ResultVO import ResultVO, FAIL_CODE, SUCCESS, DOING, SUCCESS_CODE
from client import ComputeExecutionClient
from schedule.utils.log import TaskLogFileAppender


class AbstractAsyncServiceBatchCE(ITaskHandler, IBatchOutputer, ICacheAbleHandler, IAsyncAble):
    """
    调用 异步服务的Batch处理器
    """

    # 由于这个的实现有可能是 单例的,所以要使用ThreadLocal
    def __init__(self):
        super().__init__()
        self.__code_local_host = threading.local()
        self.__process_id_local = threading.local()
        self.__execution_message_local = threading.local()

        self.__tasks_local_host = threading.local()

        self.killed = False

    def execute(self, params):
        if params is None or len(params) == 0 or ("site_" not in params) or ("projectCode_" not in params) or (
                "code_" not in params):
            # 不存在crawl的信息,没法执行
            return ResultVO(FAIL_CODE, "执行失败,没有ce的信息")

        self.__tasks_local_host.tasks = []

        output_stream = Queue()

        self.killed = False

        # 找到对应的ce
        data_compute_execution = ComputeExecutionClient.get(params["site_"], params["projectCode_"],
                                                            params["code_"])

        if data_compute_execution is None:
            return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{params['site_']}的ce")

        try:
            # 这里构造的其实就是 与执行器交互的 协议
            self.set_local_code(
                data_compute_execution.site + "_" + data_compute_execution.projectCode + "_" + data_compute_execution.code)
            self.set_process_id(params.get("__processLogId__"))
            TaskLogFileAppender.prepare_to_log(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),
                                               params["__logId__"])
            request = self.generate_request(FdInputPipe(data_compute_execution.sourceFdCodes), params)
        finally:
            self.clear_local_code()
            self.clear_process_id()
            TaskLogFileAppender.end_log()

        if request is None:
            # 返回为空,直接就返回了
            return SUCCESS

        jobId = params.get("__jobId__")
        taskLogId = params.get("__logId__")

        if request.send_type is not None and request.send_type == SendType.NOT_NEED_SEND:
            # 如果发送类型不为空,并且是不需要真正的发送出去 ,直接本地调用
            resultVO = self.do_finish(taskLogId, request.body, request.header, params, 3)
            resultVO.msg = "success with no send to actor"
            return resultVO
        else:
            # 需要发送任务给执行器,现在同步和异步任务都是一样的.
            # 发送请求
            # region 补充头
            import config
            headers = {
                'Accept': 'application/json' if request.accept_type == AcceptType.APPLICATION_JSON else 'text/plain',
                'Reply-To': (f"http://{config.server_ip or config.find_host_ip()}:{config.server_port or 5000}/async/response/{jobId}/{taskLogId}"
                             + ('' if request.accept_type == AcceptType.APPLICATION_JSON else '/plain')),
                'soap-reply-action': 'POST'
            }
            if request.header is not None:
                headers.update(request.header)
            # endregion

            global r
            try:
                if "GET" == request.method:
                    r = requests.get(request.url, headers=headers)
                elif "POST" == request.method:
                    if request.body is None:
                        r = requests.post(request.url, headers=headers)
                    else:
                        r = requests.post(request.url, json=request.body, headers=headers)
                elif "PATCH" == request.method:
                    if request.body is None:
                        r = requests.patch(request.url, headers=headers)
                    else:
                        r = requests.patch(request.url, data=request.body, headers=headers)
                elif "PUT" == request.method:
                    if request.body is None:
                        r = requests.put(request.url, headers=headers)
                    else:
                        r = requests.put(request.url, data=request.body, headers=headers)
                elif "DELETE" == request.method:
                    r = requests.delete(request.url, headers=headers)
                elif "HEAD" == request.method:
                    r = requests.head(request.url, headers=headers)
            except Exception as e:
                return ResultVO(FAIL_CODE, f"任务:{taskLogId} 请求{request.url} 失败,请求出错: {str(e)}")

            if r.status_code != 200 or r.content is None:
                return ResultVO(FAIL_CODE, f"任务:{taskLogId} 请求{request.url} 失败,请求出错: {r.content}")

        return DOING

    def __action(self, response, header: dict, task_status: int, data_compute_execution, output_stream, params,
                 is_finished, messages):
        """
        异步线程来对数据进行输入
        :param data_compute_execution:
        :param output_stream:
        :param params:
        :param is_finished:
        :param messages:
        :return:
        """
        try:
            self.set_local_code(
                data_compute_execution.site + "_" + data_compute_execution.projectCode + "_" + data_compute_execution.code)
            self.set_process_id(params.get("__processLogId__"))
            self.__execution_message_local.content = []
            TaskLogFileAppender.prepare_to_log(datetime.strptime(params["__triggerTime__"], "%Y-%m-%d %H:%M:%S"),
                                               params["__logId__"])
            messages[0] = self.do_compute(response, header, task_status,
                                          BatchOutputPipe(output_stream, len(data_compute_execution.outputFdCodes)),
                                          FdInputPipe(data_compute_execution.sourceFdCodes), params)
        finally:
            # 清理 MessageLocal 和 LocalCode
            is_finished[0] = True

            self.clear_local_code()
            self.clear_process_id()
            self.__execution_message_local.content = None
            TaskLogFileAppender.end_log()

    def do_finish(self, task_log_id: int, response, header: dict, ficus_param: dict, task_status: int) -> ResultVO:
        """
        完成的回调处理
        :param task_log_id:
        :param response:
        :param header:
        :param ficus_param:
        :param task_status:
        :return:
        """
        if ficus_param is None or len(ficus_param) == 0 or ("site_" not in ficus_param) or (
                "projectCode_" not in ficus_param) or (
                "code_" not in ficus_param):
            # 不存在crawl的信息,没法执行
            self.task_logger.error(Exception(f"执行失败,没有Ce的信息,taskId:{task_log_id}"))
            return ResultVO(FAIL_CODE, "执行失败,没有Ce的信息")

        self.__tasks_local_host.tasks = []
        output_stream = Queue()

        with ThreadPoolExecutor(max_workers=2, thread_name_prefix="async-batch-ce-") as executor:
            try:
                data_compute_execution = ComputeExecutionClient.get(ficus_param["site_"], ficus_param["projectCode_"],
                                                                    ficus_param["code_"])
                if data_compute_execution is None:
                    return ResultVO(FAIL_CODE, f"执行失败,没有找到Code:{ficus_param['code_']}的ce")

                is_finished = [False]
                messages = [None]

                self.__tasks_local_host.tasks.append(
                    executor.submit(self.__action, response, header, task_status, data_compute_execution, output_stream,
                                    ficus_param, is_finished, messages))

                # 增加发送结果的线程
                self.__tasks_local_host.tasks.append(
                    executor.submit(self.batch_output, ficus_param["code_"], data_compute_execution.outputFdCodes,
                                    is_finished,
                                    output_stream,ficus_param))
                # 阻塞主线程
                for future in as_completed(self.__tasks_local_host.tasks):
                    try:
                        data = future.result()
                    except Exception as e:
                        return ResultVO(FAIL_CODE, f"执行失败,Code:{ficus_param['code_']}的CE,原因:{str(e)}")

                if messages[0] is not None:
                    return ResultVO(SUCCESS_CODE, messages[0])
                else:
                    return DOING
            finally:
                self.__tasks_local_host.tasks.clear()
                executor.shutdown(wait=True)
                output_stream.queue.clear()

    # region 固定方法相关
    def get_execution_message_cache(self):
        """
                返回message_cache
                :return:
                """
        return self.__execution_message_local.content

    def get_batch_size(self) -> int:
        # TODO 先写死100
        return 100

    def get_code_thread_local(self):
        """
                实现上下文的code
                :return:
                """
        return self.__code_local_host

    def get_process_thread_local(self):
        """
                上下文的Id
                :return:
                """
        return self.__process_id_local

    def kill(self):
        self.killed = True
        if self.__tasks_local_host.tasks is not None:
            for task in self.__tasks_local_host.tasks:
                try:
                    task._stop()
                except:
                    pass

    def is_killed(self):
        return self.killed

    # endregion

    # region 需要实现的抽象方法
    @abstractmethod
    def generate_request(self, input_pipe: FdInputPipe, params: dict) -> AsyncServiceRequest:
        """
        生成异步任务的请求的抽象方法
        :param input_pipe:
        :param params:
        :return:
        """
        pass

    @abstractmethod
    def do_compute(self, response, header: dict, task_status: int, output_stream: BatchOutputPipe,
                   source_fds: FdInputPipe, ficus_param: dict) -> str:
        """
        异步任务的返回
        :param response:
        :param header:
        :param task_status:
        :param output_stream:
        :param source_fds:
        :param ficus_param:
        :return:
        """
        pass
    # endregion



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
# Author: lvbiao
# Created on 2019/1/22
import threading
from abc import abstractmethod

from api.exceptions import IllegalArgumentException
from api.model.FactDatasource import FactDatasourceTypeEnum
from factdatasource.execptions import CustomerDaoContextNotFoundException


class DatasourceListener(object):
    """
    用于监听数据源变化后的动作的
    """

    @abstractmethod
    def get_datasource_type(self) -> FactDatasourceTypeEnum:
        """
        获取数据源类型
        :return:
        """

    @abstractmethod
    def add_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        添加一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """

    @abstractmethod
    def update_datasource_type(self, source_name: str, url: str, credentials: str):
        """
        修改一个数据源
        :param source_name:
        :param url:
        :param credentials:
        :return:
        """

    @abstractmethod
    def delete_datasource_type(self, source_name: str):
        """
        删除一个数据源
        :param source_name:
        :return:
        """


class Datasource(object):
    """基础数据源操作"""

    @abstractmethod
    def get_client(self):
        """
        获取数据库操作的client
        :return:
        """

    @abstractmethod
    def close_client(self):
        """
        关闭客户端，关闭连接
        :return:
        """


class MultipleBaseDatasource(Datasource):
    """封装数据源的基本操作"""

    @abstractmethod
    def get_data_source(self):
        """
        获取数据源的基本信息
        :return:
        """

    def determine_current_lookup_key(self):
        """
        获取当前线程操作的数据源
        :return:
        """
        source_name = customer_dao_context_holder.get_source()
        if source_name is None:
            raise CustomerDaoContextNotFoundException('未设置上下文的数据源，无法执行操作。')
        return source_name


class BaseDatasource(Datasource):

    def init(self, source_name, url, credentials):
        if not source_name or not url:
            raise IllegalArgumentException(f'创建数据源失败，参数错误：{source_name},{url}')

        self._url = url
        self._credentials = credentials

        self._source = set()
        self._source.add(source_name)

    @abstractmethod
    def start(self):
        """分析url并开始连接数据库,设置客户端"""

    @abstractmethod
    def get_client(self):
        """
        获取数据操作的客户端
        :return:
        """

    @abstractmethod
    def close_client(self):
        """
        关闭客户端，关闭连接
        :return:
        """

    @property
    def url(self):
        return self._url

    @property
    def credentials(self):
        return self._credentials

    def add_source_name(self, ref_source_name):
        """
        当多个source使用同一个连接时，调用该方法，将该source纳入管理
        :param ref_source_name:
        :return:
        """
        self._source.add(ref_source_name)

    def remove_source_name(self, ref_source_name):
        """
        删除引用
        :param ref_source_name:
        :return:
        """
        self._source.remove(ref_source_name)

    def only_one_source(self):
        """判断该数据源是否只被一个FD享用"""
        return len(self._source) == 1


class CustomerDaoContextHolder(object):
    """
    线程的上下文,用于切换数据源. Key是系统的名字
    """

    def __init__(self):
        self.__context_holder = threading.local()
        # 由于需要在不同的线程中来判断某一个数据源是否正在使用，所以这里个添加一个变量来记录正在使用的数据源
        self.__global_context = list()

    def set_source(self, source_name: str):
        self.__context_holder.source_name = source_name
        self.__global_context.append(source_name)

    def get_source(self):
        try:
            return self.__context_holder.source_name
        except AttributeError as error:
            return None

    def clear_source(self):
        self.__global_context.remove(self.__context_holder.source_name)
        self.__context_holder.source_name = None

    def is_using(self, source_name: str):
        return source_name in self.__global_context


customer_dao_context_holder = CustomerDaoContextHolder()





class GraphNode(dict):
    """
    图数据的节点对象
    """

    def __init__(self, guid: str, type: str, tags: dict = None) -> None:
        """
        图的节点对象
        :param guid: 节点唯一guid
        :param type: 节点类型
        :param tags: 辅助标签/属性
        """
        super().__init__()
        self.__setitem__("@type", "com.sobey.jcg.sobeycube.core.api.domain.graph.GraphNode")
        self.__setitem__("guid", guid)
        self.__setitem__("type", type)
        if tags is not None:
            self.__setitem__("tags", tags)

    @property
    def guid(self):
        return self.__getitem__("guid")

    @guid.setter
    def guid(self,value:str):
        self.__setitem__("guid", value)

    @property
    def type(self):
        return self.__getitem__("type")

    @type.setter
    def type(self, value: str):
        self.__setitem__("type", value)

    @property
    def tags(self):
        return self.__getitem__("tags") if "tags" in self else None

    @tags.setter
    def tags(self, value: dict):
        if value is not None:
            self.__setitem__("tags",value)
        else:
            self.__delitem__("tags")